#!/usr/bin/env python3
from pathlib import Path
from queue import Queue
from typing import Dict, Optional, Union
from urllib.parse import urlparse

import anyio
from telegram import PhotoSize, Sticker, StickerSet
from telegram.error import TelegramError
from telegram.ext import AIORateLimiter, ApplicationBuilder

from sticker_convert.downloaders.download_base import DownloadBase
from sticker_convert.job_option import CredOption
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.metadata_handler import MetadataHandler


class DownloadTelegram(DownloadBase):
    # def __init__(self, *args: Any, **kwargs: Any) -> None:
    #     super().__init__(*args, **kwargs)

    def download_stickers_telegram(self) -> bool:
        self.token = ""

        if self.opt_cred:
            self.token = self.opt_cred.telegram_token.strip()
        if not self.token:
            self.cb.put("Download failed: Token required for downloading from telegram")
            return False

        if not ("telegram.me" in self.url or "t.me" in self.url):
            self.cb.put("Download failed: Unrecognized URL format")
            return False

        self.title = Path(urlparse(self.url).path).name

        self.emoji_dict: Dict[str, str] = {}

        return anyio.run(self.save_stickers)

    async def save_stickers(self) -> bool:
        timeout = 30
        application = (  # type: ignore
            ApplicationBuilder()
            .token(self.token)
            .rate_limiter(AIORateLimiter())
            .connect_timeout(timeout)
            .pool_timeout(timeout)
            .read_timeout(timeout)
            .write_timeout(timeout)
            .connection_pool_size(20)
            .build()
        )

        async with application:
            bot = application.bot
            try:
                sticker_set: StickerSet = await bot.get_sticker_set(
                    self.title,
                    read_timeout=timeout,
                    write_timeout=timeout,
                    connect_timeout=timeout,
                    pool_timeout=timeout,
                )
            except TelegramError as e:
                self.cb.put(
                    f"Failed to download telegram sticker set {self.title} due to: {e}"
                )
                return False

            self.cb.put(
                (
                    "bar",
                    None,
                    {
                        "set_progress_mode": "determinate",
                        "steps": len(sticker_set.stickers),
                    },
                )
            )

            async def download_sticker(
                sticker: Union[PhotoSize, Sticker], f_id: str
            ) -> None:
                sticker_file = await sticker.get_file(
                    read_timeout=timeout,
                    write_timeout=timeout,
                    connect_timeout=timeout,
                    pool_timeout=timeout,
                )
                fpath = sticker_file.file_path
                assert fpath is not None
                ext = Path(fpath).suffix
                f_name = f_id + ext
                f_path = Path(self.out_dir, f_name)
                await sticker_file.download_to_drive(
                    custom_path=f_path,
                    read_timeout=timeout,
                    write_timeout=timeout,
                    connect_timeout=timeout,
                    pool_timeout=timeout,
                )
                if isinstance(sticker, Sticker) and sticker.emoji is not None:
                    self.emoji_dict[f_id] = sticker.emoji
                self.cb.put(f"Downloaded {f_name}")
                if f_id != "cover":
                    self.cb.put("update_bar")

            async with anyio.create_task_group() as tg:
                for num, sticker in enumerate(sticker_set.stickers):
                    f_id = str(num).zfill(3)
                    tg.start_soon(download_sticker, sticker, f_id)

                if sticker_set.thumbnail is not None:
                    tg.start_soon(download_sticker, sticker_set.thumbnail, "cover")

        MetadataHandler.set_metadata(
            self.out_dir, title=self.title, emoji_dict=self.emoji_dict
        )
        return True

    @staticmethod
    def start(
        url: str,
        out_dir: Path,
        opt_cred: Optional[CredOption],
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> bool:
        downloader = DownloadTelegram(url, out_dir, opt_cred, cb, cb_return)
        return downloader.download_stickers_telegram()



#!/usr/bin/env python3
"""
Required for pack_id < 775 (https://github.com/star-39/moe-sticker-bot/blob/ef2f06f3eb7a833e011e0a5201e007fc130978e7/pkg/msbimport/import_line.go#L68)
Reference: https://stackoverflow.com/a/53622211
"""

import struct
import zlib


class ApplePngNormalize:
    @staticmethod
    def normalize(old_png: bytes) -> bytes:
        png_header = b"\x89PNG\r\n\x1a\n"

        if old_png[:8] != png_header:
            return old_png

        new_png = old_png[:8]

        chunk_pos = len(new_png)
        chunk_d = bytearray()

        found_cgbi = False

        # For each chunk in the PNG file
        width = None
        height = None
        while chunk_pos < len(old_png):
            # Reading chunk
            chunk_length_bytes = old_png[chunk_pos : chunk_pos + 4]
            chunk_length = int(struct.unpack(">L", chunk_length_bytes)[0])
            chunk_type = old_png[chunk_pos + 4 : chunk_pos + 8]
            chunk_data = old_png[chunk_pos + 8 : chunk_pos + 8 + chunk_length]
            chunk_crc = old_png[
                chunk_pos + chunk_length + 8 : chunk_pos + chunk_length + 12
            ]
            chunk_crc = struct.unpack(">L", chunk_crc)[0]
            chunk_pos += chunk_length + 12

            # Parsing the header chunk
            if chunk_type == b"IHDR":
                width = struct.unpack(">L", chunk_data[0:4])[0]
                height = struct.unpack(">L", chunk_data[4:8])[0]

            # Parsing the image chunk
            if chunk_type == b"IDAT":
                # Concatename all image data chunks
                chunk_d += chunk_data
                continue

            # Stopping the PNG file parsing
            if chunk_type == b"IEND":
                if not found_cgbi:
                    return old_png

                assert width
                assert height
                buf_size = width * height * 4 + height
                chunk_data = zlib.decompress(chunk_d, -8, buf_size)

                # Swapping red & blue bytes for each pixel
                chunk_data = bytearray(chunk_data)
                offset = 1
                for _ in range(height):
                    for x in range(width):
                        chunk_data[offset + 4 * x], chunk_data[offset + 4 * x + 2] = (
                            chunk_data[offset + 4 * x + 2],
                            chunk_data[offset + 4 * x],
                        )
                    offset += 1 + 4 * width

                # Compressing the image chunk
                # chunk_data = newdata
                chunk_data = zlib.compress(chunk_data)
                chunk_length = len(chunk_data)
                chunk_crc_idat = zlib.crc32(b"IDAT")
                chunk_crc_data = zlib.crc32(chunk_data, chunk_crc_idat)
                chunk_crc_mod = (chunk_crc_data + 0x100000000) % 0x100000000

                new_png += struct.pack(">L", chunk_length)
                new_png += b"IDAT"
                new_png += chunk_data
                new_png += struct.pack(">L", chunk_crc_mod)

                chunk_crc_type = zlib.crc32(chunk_type)
                new_png += struct.pack(">L", 0)
                new_png += b"IEND"
                new_png += struct.pack(">L", chunk_crc_type)
                break

            # Removing CgBI chunk
            if chunk_type == b"CgBI":
                found_cgbi = True
            else:
                new_png += struct.pack(">L", chunk_length)
                new_png += chunk_type
                if chunk_length > 0:
                    new_png += chunk_data
                new_png += struct.pack(">L", chunk_crc)

        return new_png



#!/usr/bin/env python3
import os
import platform
import signal
import sys
from functools import partial
from json.decoder import JSONDecodeError
from math import ceil
from multiprocessing import Event, cpu_count
from pathlib import Path
from threading import Lock, Thread
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import urlparse

from PIL import ImageFont
from ttkbootstrap import BooleanVar, DoubleVar, IntVar, StringVar, Toplevel, Window  # type: ignore
from ttkbootstrap.dialogs import Messagebox, Querybox  # type: ignore

from sticker_convert.definitions import CONFIG_DIR, DEFAULT_DIR, ROOT_DIR
from sticker_convert.gui_components.frames.comp_frame import CompFrame
from sticker_convert.gui_components.frames.config_frame import ConfigFrame
from sticker_convert.gui_components.frames.control_frame import ControlFrame
from sticker_convert.gui_components.frames.cred_frame import CredFrame
from sticker_convert.gui_components.frames.input_frame import InputFrame
from sticker_convert.gui_components.frames.output_frame import OutputFrame
from sticker_convert.gui_components.frames.progress_frame import ProgressFrame
from sticker_convert.gui_components.gui_utils import GUIUtils
from sticker_convert.job import Job
from sticker_convert.job_option import CompOption, CredOption, InputOption, OutputOption
from sticker_convert.utils.files.json_manager import JsonManager
from sticker_convert.utils.files.metadata_handler import MetadataHandler
from sticker_convert.utils.url_detect import UrlDetect
from sticker_convert.version import __version__


class GUI(Window):
    def __init__(self) -> None:
        super().__init__(themename="darkly", alpha=0)  # type: ignore
        self.init_done = False
        self.load_jsons()

        font_path = ROOT_DIR / "resources/NotoColorEmoji.ttf"
        self.emoji_font = ImageFont.truetype(font_path.as_posix(), 109)

        GUIUtils.set_icon(self)

        self.title(f"sticker-convert {__version__}")
        self.protocol("WM_DELETE_WINDOW", self.quit)

        (
            self.main_frame,
            self.horizontal_scrollbar_frame,
            self.canvas,
            self.x_scrollbar,
            self.y_scrollbar,
            self.scrollable_frame,
        ) = GUIUtils.create_scrollable_frame(self)

        self.declare_variables()
        self.apply_config()
        self.apply_creds()
        self.init_frames()
        self.pack_frames()
        self.warn_tkinter_bug()
        GUIUtils.finalize_window(self)

        self.bind("<<exec_in_main>>", self.exec_in_main)  # type: ignore

    def __enter__(self) -> "GUI":
        return self

    def gui(self) -> None:
        self.init_done = True
        self.highlight_fields()
        self.mainloop()

    def quit(self) -> None:
        if self.job:
            response = self.cb_ask_bool("Job is running, really quit?")
            if response is False:
                return

        self.cb_msg(msg="Quitting, please wait...")

        self.save_config()
        if self.settings_save_cred_var.get() is True:
            self.save_creds()
        else:
            self.delete_creds()

        if self.job:
            self.cancel_job()
        self.destroy()

    def declare_variables(self) -> None:
        # Input
        self.input_option_display_var = StringVar(self)
        self.input_option_true_var = StringVar(self)
        self.input_setdir_var = StringVar(self)
        self.input_address_var = StringVar(self)

        # Compression
        self.no_compress_var = BooleanVar()
        self.comp_preset_var = StringVar(self)
        self.fps_min_var = IntVar(self)
        self.fps_max_var = IntVar(self)
        self.fps_disable_var = BooleanVar()
        self.fps_power_var = DoubleVar()
        self.res_w_min_var = IntVar(self)
        self.res_w_max_var = IntVar(self)
        self.res_w_disable_var = BooleanVar()
        self.res_h_min_var = IntVar(self)
        self.res_h_max_var = IntVar(self)
        self.res_h_disable_var = BooleanVar()
        self.res_power_var = DoubleVar()
        self.quality_min_var = IntVar(self)
        self.quality_max_var = IntVar(self)
        self.quality_disable_var = BooleanVar()
        self.quality_power_var = DoubleVar()
        self.color_min_var = IntVar(self)
        self.color_max_var = IntVar(self)
        self.color_disable_var = BooleanVar()
        self.color_power_var = DoubleVar()
        self.duration_min_var = IntVar(self)
        self.duration_max_var = IntVar(self)
        self.duration_disable_var = BooleanVar()
        self.img_size_max_var = IntVar(self)
        self.vid_size_max_var = IntVar(self)
        self.size_disable_var = BooleanVar()
        self.img_format_var = StringVar(self)
        self.vid_format_var = StringVar(self)
        self.fake_vid_var = BooleanVar()
        self.scale_filter_var = StringVar(self)
        self.quantize_method_var = StringVar(self)
        self.cache_dir_var = StringVar(self)
        self.default_emoji_var = StringVar(self)
        self.steps_var = IntVar(self)
        self.processes_var = IntVar(self)

        # Output
        self.output_option_display_var = StringVar(self)
        self.output_option_true_var = StringVar(self)
        self.output_setdir_var = StringVar(self)
        self.title_var = StringVar(self)
        self.author_var = StringVar(self)

        # Credentials
        self.signal_uuid_var = StringVar(self)
        self.signal_password_var = StringVar(self)
        self.signal_data_dir_var = StringVar(self)
        self.telegram_token_var = StringVar(self)
        self.telegram_userid_var = StringVar(self)
        self.kakao_auth_token_var = StringVar(self)
        self.kakao_username_var = StringVar(self)
        self.kakao_password_var = StringVar(self)
        self.kakao_country_code_var = StringVar(self)
        self.kakao_phone_number_var = StringVar(self)
        self.line_cookies_var = StringVar(self)

        # Config
        self.settings_save_cred_var = BooleanVar()

        # Other
        self.msg_lock = Lock()
        self.bar_lock = Lock()
        self.response_event = Event()
        self.response = None
        self.action: Optional[Callable[..., Any]] = None
        self.job: Optional[Job] = None

    def init_frames(self) -> None:
        self.input_frame = InputFrame(
            self, self.scrollable_frame, borderwidth=1, text="Input"
        )
        self.comp_frame = CompFrame(
            self, self.scrollable_frame, borderwidth=1, text="Compression options"
        )
        self.output_frame = OutputFrame(
            self, self.scrollable_frame, borderwidth=1, text="Output"
        )
        self.cred_frame = CredFrame(
            self, self.scrollable_frame, borderwidth=1, text="Credentials"
        )
        self.settings_frame = ConfigFrame(
            self, self.scrollable_frame, borderwidth=1, text="Config"
        )
        self.progress_frame = ProgressFrame(
            self, self.scrollable_frame, borderwidth=1, text="Progress"
        )
        self.control_frame = ControlFrame(self, self.scrollable_frame, borderwidth=1)

    def pack_frames(self) -> None:
        self.input_frame.grid(column=0, row=0, sticky="w", padx=5, pady=5)
        self.comp_frame.grid(column=1, row=0, sticky="news", padx=5, pady=5)
        self.output_frame.grid(column=0, row=1, sticky="w", padx=5, pady=5)
        self.cred_frame.grid(column=1, row=1, rowspan=2, sticky="w", padx=5, pady=5)
        self.settings_frame.grid(column=0, row=2, sticky="news", padx=5, pady=5)
        self.progress_frame.grid(
            column=0, row=3, columnspan=2, sticky="news", padx=5, pady=5
        )
        self.control_frame.grid(
            column=0, row=4, columnspan=2, sticky="news", padx=5, pady=5
        )

    def warn_tkinter_bug(self) -> None:
        if (
            platform.system() == "Darwin"
            and platform.mac_ver()[0].split(".")[0] == "14"
            and sys.version_info[0] == 3
            and sys.version_info[1] == 11
            and sys.version_info[2] <= 6
        ):
            msg = "NOTICE: If buttons are not responsive, try to press "
            msg += "on title bar or move mouse cursor away from window for a while."
            self.cb_msg(msg)
            msg = (
                "(This is due to a bug in tkinter specific to macOS 14 python <=3.11.6)"
            )
            self.cb_msg(msg)
            msg = "(https://github.com/python/cpython/issues/110218)"
            self.cb_msg(msg)

    def load_jsons(self) -> None:
        try:
            from sticker_convert.utils.files.json_resources_loader import COMPRESSION_JSON, EMOJI_JSON, HELP_JSON, INPUT_JSON, OUTPUT_JSON
        except RuntimeError as e:
            self.cb_msg(str(e))
            return

        self.help = HELP_JSON
        self.input_presets = INPUT_JSON
        self.compression_presets = COMPRESSION_JSON
        self.output_presets = OUTPUT_JSON
        self.emoji_list = EMOJI_JSON

        if not (
            self.compression_presets and self.input_presets and self.output_presets
        ):
            Messagebox.show_error(  # type: ignore
                message='Warning: json(s) under "resources" directory cannot be found',
                title="sticker-convert",
            )
            sys.exit()

        self.settings_path = CONFIG_DIR / "config.json"
        if self.settings_path.is_file():
            try:
                self.settings: Dict[Any, Any] = JsonManager.load_json(
                    self.settings_path
                )
            except JSONDecodeError:
                self.cb_msg("Warning: config.json content is corrupted")
                self.settings = {}
        else:
            self.settings = {}

        self.creds_path = CONFIG_DIR / "creds.json"
        if self.creds_path.is_file():
            try:
                self.creds = JsonManager.load_json(self.creds_path)
            except JSONDecodeError:
                self.cb_msg("Warning: creds.json content is corrupted")
                self.creds = {}
        else:
            self.creds = {}

    def save_config(self) -> None:
        # Only update comp_custom if custom preset is selected
        if self.comp_preset_var.get() == "custom":
            comp_custom = self.get_opt_comp().to_dict()
            del comp_custom["preset"]
            del comp_custom["no_compress"]
        else:
            compression_presets_custom = self.compression_presets.get("custom")
            if compression_presets_custom is None:
                comp_custom = {}
            else:
                comp_custom = compression_presets_custom

        self.settings = {
            "input": self.get_opt_input().to_dict(),
            "comp": {
                "no_compress": self.no_compress_var.get(),
                "preset": self.comp_preset_var.get(),
                "cache_dir": self.cache_dir_var.get(),
                "processes": self.processes_var.get(),
            },
            "comp_custom": comp_custom,
            "output": self.get_opt_output().to_dict(),
            "creds": {"save_cred": self.settings_save_cred_var.get()},
        }

        JsonManager.save_json(self.settings_path, self.settings)

    def save_creds(self) -> None:
        self.creds = self.get_opt_cred().to_dict()

        JsonManager.save_json(self.creds_path, self.creds)

    def delete_creds(self) -> None:
        if self.creds_path.is_file():
            os.remove(self.creds_path)

    def delete_config(self) -> None:
        if self.settings_path.is_file():
            os.remove(self.settings_path)

    def apply_config(self) -> None:
        # Input
        self.default_input_mode: str = self.settings.get("input", {}).get(
            "option", "auto"
        )
        self.input_address_var.set(self.settings.get("input", {}).get("url", ""))
        default_stickers_input_dir = str(DEFAULT_DIR / "stickers_input")
        self.input_setdir_var.set(
            self.settings.get("input", {}).get("dir", default_stickers_input_dir)
        )
        if not Path(self.input_setdir_var.get()).is_dir():
            self.input_setdir_var.set(default_stickers_input_dir)
        self.input_option_display_var.set(
            self.input_presets[self.default_input_mode]["full_name"]
        )
        self.input_option_true_var.set(
            self.input_presets[self.default_input_mode]["full_name"]
        )

        # Compression
        self.no_compress_var.set(
            self.settings.get("comp", {}).get("no_compress", False)
        )
        default_comp_preset = list(self.compression_presets.keys())[0]
        self.comp_preset_var.set(
            self.settings.get("comp", {}).get("preset", default_comp_preset)
        )
        comp_custom = self.settings.get("comp_custom")
        if comp_custom:
            self.compression_presets["custom"] = comp_custom
        self.cache_dir_var.set(self.settings.get("comp", {}).get("cache_dir", ""))
        self.processes_var.set(
            self.settings.get("comp", {}).get("processes", ceil(cpu_count() / 2))
        )
        self.default_output_mode: str = self.settings.get("output", {}).get(
            "option", "signal"
        )

        # Output
        default_stickers_output_dir = str(DEFAULT_DIR / "stickers_output")
        self.output_setdir_var.set(
            self.settings.get("output", {}).get("dir", default_stickers_output_dir)
        )
        if not Path(self.output_setdir_var.get()).is_dir():
            self.output_setdir_var.set(default_stickers_output_dir)
        self.title_var.set(self.settings.get("output", {}).get("title", ""))
        self.author_var.set(self.settings.get("output", {}).get("author", ""))
        self.settings_save_cred_var.set(
            self.settings.get("creds", {}).get("save_cred", True)
        )
        self.output_option_display_var.set(
            self.output_presets[self.default_output_mode]["full_name"]
        )
        self.output_option_true_var.set(
            self.output_presets[self.default_output_mode]["full_name"]
        )

    def apply_creds(self) -> None:
        self.signal_uuid_var.set(self.creds.get("signal", {}).get("uuid", ""))
        self.signal_password_var.set(self.creds.get("signal", {}).get("password", ""))
        self.telegram_token_var.set(self.creds.get("telegram", {}).get("token", ""))
        self.telegram_userid_var.set(self.creds.get("telegram", {}).get("userid", ""))
        self.kakao_auth_token_var.set(self.creds.get("kakao", {}).get("auth_token", ""))
        self.kakao_username_var.set(self.creds.get("kakao", {}).get("username", ""))
        self.kakao_password_var.set(self.creds.get("kakao", {}).get("password", ""))
        self.kakao_country_code_var.set(
            self.creds.get("kakao", {}).get("country_code", "")
        )
        self.kakao_phone_number_var.set(
            self.creds.get("kakao", {}).get("phone_number", "")
        )
        self.line_cookies_var.set(self.creds.get("line", {}).get("cookies", ""))

    def get_input_name(self) -> str:
        return [
            k
            for k, v in self.input_presets.items()
            if v["full_name"] == self.input_option_true_var.get()
        ][0]

    def get_input_display_name(self) -> str:
        return [
            k
            for k, v in self.input_presets.items()
            if v["full_name"] == self.input_option_display_var.get()
        ][0]

    def get_output_name(self) -> str:
        return [
            k
            for k, v in self.output_presets.items()
            if v["full_name"] == self.output_option_true_var.get()
        ][0]

    # def get_output_display_name(self) -> str:
    #     return [
    #         k
    #         for k, v in self.output_presets.items()
    #         if v["full_name"] == self.output_option_display_var.get()
    #     ][0]

    def get_preset(self) -> str:
        selection = self.comp_preset_var.get()
        if selection == "auto":
            output_option = self.get_output_name()
            if output_option == "imessage":
                return "imessage_small"
            if output_option == "local":
                return selection
            return output_option

        return selection

    def start_job(self) -> None:
        self.save_config()
        if self.settings_save_cred_var.get() is True:
            self.save_creds()
        else:
            self.delete_creds()

        self.control_frame.start_btn.config(text="Cancel", bootstyle="danger")  # type: ignore
        self.set_inputs("disabled")

        opt_input = self.get_opt_input()
        opt_output = self.get_opt_output()
        opt_comp = self.get_opt_comp()
        opt_cred = self.get_opt_cred()

        self.job = Job(
            opt_input,
            opt_comp,
            opt_output,
            opt_cred,
            self.cb_msg,
            self.cb_msg_block,
            self.cb_bar,
            self.cb_ask_bool,
            self.cb_ask_str,
        )

        signal.signal(signal.SIGINT, self.job.cancel)

        Thread(target=self.start_process, daemon=True).start()

    def get_opt_input(self) -> InputOption:
        return InputOption(
            option=self.get_input_name(),
            url=self.input_address_var.get(),
            dir=Path(self.input_setdir_var.get()),
        )

    def get_opt_output(self) -> OutputOption:
        return OutputOption(
            option=self.get_output_name(),
            dir=Path(self.output_setdir_var.get()),
            title=self.title_var.get(),
            author=self.author_var.get(),
        )

    def get_opt_comp(self) -> CompOption:
        return CompOption(
            preset=self.get_preset(),
            size_max_img=self.img_size_max_var.get()
            if not self.size_disable_var.get()
            else None,
            size_max_vid=self.vid_size_max_var.get()
            if not self.size_disable_var.get()
            else None,
            format_img=(self.img_format_var.get(),),
            format_vid=(self.vid_format_var.get(),),
            fps_min=self.fps_min_var.get() if not self.fps_disable_var.get() else None,
            fps_max=self.fps_max_var.get() if not self.fps_disable_var.get() else None,
            fps_power=self.fps_power_var.get(),
            res_w_min=self.res_w_min_var.get()
            if not self.res_w_disable_var.get()
            else None,
            res_w_max=self.res_w_max_var.get()
            if not self.res_w_disable_var.get()
            else None,
            res_h_min=self.res_h_min_var.get()
            if not self.res_h_disable_var.get()
            else None,
            res_h_max=self.res_h_max_var.get()
            if not self.res_h_disable_var.get()
            else None,
            res_power=self.res_power_var.get(),
            quality_min=self.quality_min_var.get()
            if not self.quality_disable_var.get()
            else None,
            quality_max=self.quality_max_var.get()
            if not self.quality_disable_var.get()
            else None,
            quality_power=self.quality_power_var.get(),
            color_min=self.color_min_var.get()
            if not self.color_disable_var.get()
            else None,
            color_max=self.color_max_var.get()
            if not self.color_disable_var.get()
            else None,
            color_power=self.color_power_var.get(),
            duration_min=self.duration_min_var.get()
            if not self.duration_disable_var.get()
            else None,
            duration_max=self.duration_max_var.get()
            if not self.duration_disable_var.get()
            else None,
            steps=self.steps_var.get(),
            fake_vid=self.fake_vid_var.get(),
            scale_filter=self.scale_filter_var.get(),
            quantize_method=self.quantize_method_var.get(),
            cache_dir=self.cache_dir_var.get()
            if self.cache_dir_var.get() != ""
            else None,
            default_emoji=self.default_emoji_var.get(),
            no_compress=self.no_compress_var.get(),
            processes=self.processes_var.get(),
        )

    def get_opt_cred(self) -> CredOption:
        return CredOption(
            signal_uuid=self.signal_uuid_var.get(),
            signal_password=self.signal_password_var.get(),
            telegram_token=self.telegram_token_var.get(),
            telegram_userid=self.telegram_userid_var.get(),
            kakao_auth_token=self.kakao_auth_token_var.get(),
            kakao_username=self.kakao_username_var.get(),
            kakao_password=self.kakao_password_var.get(),
            kakao_country_code=self.kakao_country_code_var.get(),
            kakao_phone_number=self.kakao_phone_number_var.get(),
            line_cookies=self.line_cookies_var.get(),
        )

    def start_process(self) -> None:
        if self.job:
            self.job.start()
        self.job = None

        self.stop_job()

    def stop_job(self) -> None:
        self.set_inputs("normal")
        self.control_frame.start_btn.config(text="Start", bootstyle="default")  # type: ignore

    def cancel_job(self) -> None:
        if self.job:
            self.cb_msg(msg="Cancelling job...")
            self.job.cancel()
            self.cb_bar(set_progress_mode="clear")

    def set_inputs(self, state: str) -> None:
        # state: 'normal', 'disabled'

        self.input_frame.set_states(state=state)
        self.comp_frame.set_states(state=state)
        self.output_frame.set_states(state=state)
        self.cred_frame.set_states(state=state)
        self.settings_frame.set_states(state=state)

        if state == "normal":
            self.input_frame.cb_input_option()
            self.comp_frame.cb_no_compress()

    def exec_in_main(self, _evt: Any) -> Any:
        if self.action:
            self.response = self.action()
        self.response_event.set()

    def cb_ask_str(
        self,
        question: str,
        initialvalue: Optional[str] = None,
        cli_show_initialvalue: bool = True,
        parent: Optional[object] = None,
    ) -> Any:
        self.action = partial(
            Querybox.get_string,  # type: ignore
            question,
            title="sticker-convert",
            initialvalue=initialvalue,
            parent=parent,
        )
        self.event_generate("<<exec_in_main>>")
        self.response_event.wait()
        self.response_event.clear()

        return self.response

    def cb_ask_bool(
        self, question: str, parent: Union[Window, Toplevel, None] = None
    ) -> bool:
        self.action = partial(
            Messagebox.yesno,  # type: ignore
            question,
            title="sticker-convert",
            parent=parent,
        )
        self.event_generate("<<exec_in_main>>")
        self.response_event.wait()
        self.response_event.clear()

        if self.response == "Yes":
            return True
        return False

    def cb_msg(self, *args: Any, **kwargs: Any) -> None:
        with self.msg_lock:
            self.progress_frame.update_message_box(*args, **kwargs)

    def cb_msg_block(
        self,
        *args: Any,
        message: Optional[str] = None,
        parent: Optional[object] = None,
        **_kwargs: Any,
    ) -> Any:
        if message is None and len(args) > 0:
            message = " ".join(str(i) for i in args)
        self.action = partial(
            Messagebox.show_info,  # type: ignore
            message,
            title="sticker-convert",
            parent=parent,
        )
        self.event_generate("<<exec_in_main>>")
        self.response_event.wait()
        self.response_event.clear()

        return self.response

    def cb_bar(
        self,
        *args: Any,
        set_progress_mode: Optional[str] = None,
        steps: int = 0,
        update_bar: bool = False,
        **kwargs: Any,
    ) -> None:
        with self.bar_lock:
            self.progress_frame.update_progress_bar(
                set_progress_mode, steps, update_bar, *args, **kwargs
            )

    def highlight_fields(self) -> bool:
        if not self.init_done:
            return True

        input_option = self.get_input_name()
        input_option_display = self.get_input_display_name()
        output_option = self.get_output_name()
        # output_option_display = self.get_output_display_name()
        url = self.input_address_var.get()

        in_out_dir_same = (
            Path(self.input_setdir_var.get()).absolute()
            == Path(self.output_setdir_var.get()).absolute()
        )

        # Input
        if in_out_dir_same is True:
            self.input_frame.input_setdir_entry.config(bootstyle="danger")  # type: ignore
        elif not Path(self.input_setdir_var.get()).is_dir():
            self.input_frame.input_setdir_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.input_frame.input_setdir_entry.config(bootstyle="default")  # type: ignore

        self.input_frame.address_lbl.config(
            text=self.input_presets[input_option_display]["address_lbls"]
        )
        self.input_frame.address_entry.config(bootstyle="default")  # type: ignore

        if input_option == "local":
            self.input_frame.address_entry.config(state="disabled")
            self.input_frame.address_tip.config(
                text=self.input_presets[input_option_display]["example"]
            )

        else:
            self.input_frame.address_entry.config(state="normal")
            self.input_frame.address_tip.config(
                text=self.input_presets[input_option_display]["example"]
            )
            download_option = UrlDetect.detect(url)

            if not url:
                self.input_frame.address_entry.config(bootstyle="warning")  # type: ignore

            elif download_option != input_option and not (
                input_option in ("kakao", "line") and url.isnumeric()
            ):
                self.input_frame.address_entry.config(bootstyle="danger")  # type: ignore
                self.input_frame.address_tip.config(
                    text=f"Invalid URL. {self.input_presets[input_option_display]['example']}"
                )

            elif input_option_display == "auto" and download_option:
                self.input_frame.address_tip.config(
                    text=f"Detected URL: {download_option}"
                )

        # Output
        if in_out_dir_same is True:
            self.output_frame.output_setdir_entry.config(bootstyle="danger")  # type: ignore
        elif not Path(self.output_setdir_var.get()).is_dir():
            self.output_frame.output_setdir_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.output_frame.output_setdir_entry.config(bootstyle="default")  # type: ignore

        if (
            MetadataHandler.check_metadata_required(output_option, "title")
            and not MetadataHandler.check_metadata_provided(
                Path(self.input_setdir_var.get()), input_option, "title"
            )
            and not self.title_var.get()
        ):
            self.output_frame.title_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.output_frame.title_entry.config(bootstyle="default")  # type: ignore

        if (
            MetadataHandler.check_metadata_required(output_option, "author")
            and not MetadataHandler.check_metadata_provided(
                Path(self.input_setdir_var.get()), input_option, "author"
            )
            and not self.author_var.get()
        ):
            self.output_frame.author_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.output_frame.author_entry.config(bootstyle="default")  # type: ignore

        if self.comp_preset_var.get() == "auto":
            if output_option == "local":
                self.no_compress_var.set(True)
            else:
                self.no_compress_var.set(False)
            self.comp_frame.cb_no_compress()

        # Credentials
        if output_option == "signal" and not self.signal_uuid_var.get():
            self.cred_frame.signal_uuid_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.cred_frame.signal_uuid_entry.config(bootstyle="default")  # type: ignore

        if output_option == "signal" and not self.signal_password_var.get():
            self.cred_frame.signal_password_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.cred_frame.signal_password_entry.config(bootstyle="default")  # type: ignore

        if (
            input_option == "telegram" or output_option == "telegram"
        ) and not self.telegram_token_var.get():
            self.cred_frame.telegram_token_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.cred_frame.telegram_token_entry.config(bootstyle="default")  # type: ignore

        if output_option == "telegram" and not self.telegram_userid_var.get():
            self.cred_frame.telegram_userid_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.cred_frame.telegram_userid_entry.config(bootstyle="default")  # type: ignore

        if (
            urlparse(url).netloc == "e.kakao.com"
            and not self.kakao_auth_token_var.get()
        ):
            self.cred_frame.kakao_auth_token_entry.config(bootstyle="warning")  # type: ignore
        else:
            self.cred_frame.kakao_auth_token_entry.config(bootstyle="default")  # type: ignore

        # Check for Input and Compression mismatch
        if (
            not self.no_compress_var.get()
            and self.get_output_name() != "local"
            and self.comp_preset_var.get() not in ("auto", "custom")
            and self.get_output_name() not in self.comp_preset_var.get()
        ):
            self.comp_frame.comp_preset_opt.config(bootstyle="warning")  # type: ignore
            self.output_frame.output_option_opt.config(bootstyle="warning")  # type: ignore
        else:
            self.comp_frame.comp_preset_opt.config(bootstyle="secondary")  # type: ignore
            self.output_frame.output_option_opt.config(bootstyle="secondary")  # type: ignore

        return True



#!/usr/bin/env python3
import platform
from typing import TYPE_CHECKING, Tuple

from ttkbootstrap import Toplevel  # type: ignore

from sticker_convert.gui_components.gui_utils import GUIUtils

if TYPE_CHECKING:
    from sticker_convert.gui import GUI  # type: ignore


class BaseWindow(Toplevel):
    def __init__(self, gui: "GUI") -> None:
        super().__init__(alpha=0)  # type: ignore
        self.gui = gui

        GUIUtils.set_icon(self)

        self.mousewheel: Tuple[str, ...]
        if platform.system() == "Windows":
            self.mousewheel = ("<MouseWheel>",)
            self.delta_divide = 120
        elif platform.system() == "Darwin":
            self.mousewheel = ("<MouseWheel>",)
            self.delta_divide = 1
        else:
            self.mousewheel = ("<Button-4>", "<Button-5>")
            self.delta_divide = 120

        (
            self.main_frame,
            self.horizontal_scrollbar_frame,
            self.canvas,
            self.x_scrollbar,
            self.y_scrollbar,
            self.scrollable_frame,
        ) = GUIUtils.create_scrollable_frame(self)



#!/usr/bin/env python3
from pathlib import Path
from queue import Queue
from typing import Dict, Optional, Union

import anyio
from signalstickers_client import StickersClient  # type: ignore
from signalstickers_client.errors import SignalException  # type: ignore
from signalstickers_client.models import StickerPack  # type: ignore

from sticker_convert.downloaders.download_base import DownloadBase
from sticker_convert.job_option import CredOption
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.metadata_handler import MetadataHandler
from sticker_convert.utils.media.codec_info import CodecInfo


class DownloadSignal(DownloadBase):
    # def __init__(self, *args: Any, **kwargs: Any) -> None:
    #     super().__init__(*args, **kwargs)

    @staticmethod
    async def get_pack(pack_id: str, pack_key: str) -> StickerPack:
        async with StickersClient() as client:
            pack = await client.get_pack(pack_id, pack_key)  # type: ignore

        return pack

    def save_stickers(self, pack: StickerPack) -> None:
        self.cb.put(
            (
                "bar",
                None,
                {
                    "set_progress_mode": "determinate",
                    "steps": len(pack.stickers),  # type: ignore
                },
            )
        )

        emoji_dict: Dict[str, str] = {}
        for sticker in pack.stickers:  # type: ignore
            f_id = str(sticker.id).zfill(3)  # type: ignore
            f_path = Path(self.out_dir, f_id)
            with open(f_path, "wb") as f:
                f.write(sticker.image_data)  # type: ignore

            emoji_dict[f_id] = sticker.emoji  # type: ignore

            codec = CodecInfo.get_file_codec(f_path)
            if codec == "":
                msg = f"Warning: Downloaded {f_path} but cannot get file codec"
                self.cb.put(msg)
            else:
                f_path_new = Path(f"{f_path}.{codec}")
                f_path.rename(f_path_new)
                msg = f"Downloaded {f_id}.{codec}"
                self.cb.put(msg)

            self.cb.put("update_bar")

        MetadataHandler.set_metadata(
            self.out_dir, title=pack.title, author=pack.author, emoji_dict=emoji_dict
        )

    def download_stickers_signal(self) -> bool:
        if "signal.art" not in self.url:
            self.cb.put("Download failed: Unrecognized URL format")
            return False

        pack_id = self.url.split("#pack_id=")[1].split("&pack_key=")[0]
        pack_key = self.url.split("&pack_key=")[1]

        try:
            pack = anyio.run(DownloadSignal.get_pack, pack_id, pack_key)
        except SignalException as e:
            self.cb.put(f"Failed to download pack due to {repr(e)}")
            return False

        self.save_stickers(pack)

        return True

    @staticmethod
    def start(
        url: str,
        out_dir: Path,
        opt_cred: Optional[CredOption],
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> bool:
        downloader = DownloadSignal(url, out_dir, opt_cred, cb, cb_return)
        return downloader.download_stickers_signal()



#!/usr/bin/env python3
from __future__ import annotations

import os
import platform
import shutil
import traceback
from datetime import datetime
from multiprocessing import Process, Value
from multiprocessing.managers import ListProxy, SyncManager
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Iterator, List, Optional, Tuple
from urllib.parse import urlparse

from sticker_convert.converter import StickerConvert
from sticker_convert.downloaders.download_kakao import DownloadKakao
from sticker_convert.downloaders.download_line import DownloadLine
from sticker_convert.downloaders.download_signal import DownloadSignal
from sticker_convert.downloaders.download_telegram import DownloadTelegram
from sticker_convert.job_option import CompOption, CredOption, InputOption, OutputOption
from sticker_convert.uploaders.compress_wastickers import CompressWastickers
from sticker_convert.uploaders.upload_signal import UploadSignal
from sticker_convert.uploaders.upload_telegram import UploadTelegram
from sticker_convert.uploaders.xcode_imessage import XcodeImessage
from sticker_convert.utils.callback import CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.json_resources_loader import OUTPUT_JSON
from sticker_convert.utils.files.metadata_handler import MetadataHandler
from sticker_convert.utils.media.codec_info import CodecInfo

CbQueueType = Queue[CbQueueItemType]
WorkListItemType = Optional[Tuple[Callable[..., Any], Tuple[Any, ...]]]
if TYPE_CHECKING:
    # mypy complains about this
    WorkListType = ListProxy[WorkListItemType]  # type: ignore
else:
    WorkListType = List[WorkListItemType]


class Executor:
    def __init__(
        self,
        cb_msg: Callable[..., None],
        cb_msg_block: Callable[..., None],
        cb_bar: Callable[..., None],
        cb_ask_bool: Callable[..., bool],
        cb_ask_str: Callable[..., str],
    ) -> None:
        self.cb_msg = cb_msg
        self.cb_msg_block = cb_msg_block
        self.cb_bar = cb_bar
        self.cb_ask_bool = cb_ask_bool
        self.cb_ask_str = cb_ask_str

        self.manager = SyncManager()
        self.manager.start()
        # Using list instead of queue for work_list as it can cause random deadlocks
        # Especially when using scale_filter=nearest
        self.work_list: WorkListType = self.manager.list()
        self.results_queue: Queue[Any] = self.manager.Queue()
        self.cb_queue: CbQueueType = self.manager.Queue()
        self.cb_return = CallbackReturn()
        self.processes: List[Process] = []

        self.is_cancel_job = Value("i", 0)

        self.cb_thread_instance = Thread(
            target=self.cb_thread,
            args=(
                self.cb_queue,
                self.cb_return,
            ),
        )
        self.cb_thread_instance.start()

    def cb_thread(
        self,
        cb_queue: CbQueueType,
        cb_return: CallbackReturn,
    ) -> None:
        for i in iter(cb_queue.get, None):
            if isinstance(i, tuple):
                action = i[0]
                if len(i) >= 2:
                    args: Tuple[str, ...] = i[1] if i[1] else tuple()
                else:
                    args = tuple()
                if len(i) >= 3:
                    kwargs: Dict[str, str] = i[2] if i[2] else {}
                else:
                    kwargs = {}
            else:
                action = i
                args = tuple()
                kwargs = {}
            if action == "msg":
                self.cb_msg(*args, **kwargs)
            elif action == "bar":
                self.cb_bar(*args, **kwargs)
            elif action == "update_bar":
                self.cb_bar(update_bar=True)
            elif action == "msg_block":
                cb_return.set_response(self.cb_msg_block(*args, **kwargs))
            elif action == "ask_bool":
                cb_return.set_response(self.cb_ask_bool(*args, **kwargs))
            elif action == "ask_str":
                cb_return.set_response(self.cb_ask_str(*args, **kwargs))
            else:
                self.cb_msg(action)

    @staticmethod
    def worker(
        work_list: WorkListType,
        results_queue: Queue[Any],
        cb_queue: CbQueueType,
        cb_return: CallbackReturn,
    ) -> None:
        while True:
            try:
                work = work_list.pop(0)
            except IndexError:
                break

            if work is None:
                break
            else:
                work_func, work_args = work

            try:
                results = work_func(*work_args, cb_queue, cb_return)
                results_queue.put(results)
            except Exception:
                arg_dump: List[Any] = []
                for i in work_args:
                    if isinstance(i, CredOption):
                        arg_dump.append("CredOption(REDACTED)")
                    else:
                        arg_dump.append(i)
                e = "##### EXCEPTION #####\n"
                e += "Function: " + repr(work_func) + "\n"
                e += "Arguments: " + repr(arg_dump) + "\n"
                e += traceback.format_exc()
                e += "#####################"
                cb_queue.put(e)
        work_list.append(None)

    def start_workers(self, processes: int = 1) -> None:
        for _ in range(processes):
            process = Process(
                target=Executor.worker,
                args=(
                    self.work_list,
                    self.results_queue,
                    self.cb_queue,
                    self.cb_return,
                ),
                daemon=True,
            )

            process.start()
            self.processes.append(process)

    def add_work(
        self, work_func: Callable[..., Any], work_args: Tuple[Any, ...]
    ) -> None:
        self.work_list.append((work_func, work_args))

    def join_workers(self) -> None:
        self.work_list.append(None)
        try:
            for process in self.processes:
                process.join()
        except KeyboardInterrupt:
            pass

        self.results_queue.put(None)
        self.work_list = self.manager.list()
        self.processes.clear()

    def kill_workers(self, *_: Any, **__: Any) -> None:
        self.is_cancel_job.value = 1  # type: ignore
        self.work_list = self.manager.list()

        for process in self.processes:
            if platform.system() == "Windows":
                process.terminate()
            else:
                process.close()
            process.join()

        self.cleanup()

    def cleanup(self) -> None:
        self.cb_queue.put(None)
        self.cb_thread_instance.join()
        self.work_list = self.manager.list()

    def get_result(self) -> Generator[Any, None, None]:
        gen: Iterator[Any] = iter(self.results_queue.get, None)
        yield from gen

    def cb(
        self,
        action: Optional[str],
        args: Optional[Tuple[str, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.cb_queue.put((action, args, kwargs))


class Job:
    def __init__(
        self,
        opt_input: InputOption,
        opt_comp: CompOption,
        opt_output: OutputOption,
        opt_cred: CredOption,
        cb_msg: Callable[..., None],
        cb_msg_block: Callable[..., None],
        cb_bar: Callable[..., None],
        cb_ask_bool: Callable[..., bool],
        cb_ask_str: Callable[..., str],
    ) -> None:
        self.opt_input = opt_input
        self.opt_comp = opt_comp
        self.opt_output = opt_output
        self.opt_cred = opt_cred
        self.cb_msg = cb_msg
        self.cb_msg_block = cb_msg_block
        self.cb_bar = cb_bar
        self.cb_ask_bool = cb_ask_bool
        self.cb_ask_str = cb_ask_str

        self.compress_fails: List[str] = []
        self.out_urls: List[str] = []

        self.executor = Executor(
            self.cb_msg,
            self.cb_msg_block,
            self.cb_bar,
            self.cb_ask_bool,
            self.cb_ask_str,
        )

    def start(self) -> int:
        if Path(self.opt_input.dir).is_dir() is False:
            os.makedirs(self.opt_input.dir)

        if Path(self.opt_output.dir).is_dir() is False:
            os.makedirs(self.opt_output.dir)

        self.executor.cb("msg", kwargs={"cls": True})

        tasks = (
            self.verify_input,
            self.cleanup,
            self.download,
            self.compress,
            self.export,
            self.report,
        )

        code = 0
        for task in tasks:
            self.executor.cb("bar", kwargs={"set_progress_mode": "indeterminate"})
            success = task()

            if self.executor.is_cancel_job.value == 1:  # type: ignore
                code = 2
                self.executor.cb("Job cancelled.")
                break
            if not success:
                code = 1
                self.executor.cb("An error occured during this run.")
                break

        self.executor.cb("bar", kwargs={"set_progress_mode": "clear"})
        self.executor.cleanup()

        return code

    def cancel(self, *_: Any, **_kwargs: Any) -> None:
        self.executor.kill_workers()

    def verify_input(self) -> bool:
        info_msg = ""
        error_msg = ""

        save_to_local_tip = ""
        save_to_local_tip += "    If you want to upload the results by yourself,\n"
        save_to_local_tip += '    select "Save to local directory only" for output\n'

        if Path(self.opt_input.dir).resolve() == Path(self.opt_output.dir).resolve():
            error_msg += "\n"
            error_msg += "[X] Input and output directories cannot be the same\n"

        if self.opt_input.option == "auto":
            error_msg += "\n"
            error_msg += "[X] Unrecognized URL input source\n"

        if self.opt_input.option != "local" and not self.opt_input.url:
            error_msg += "\n"
            error_msg += "[X] URL address cannot be empty.\n"
            error_msg += "    If you only want to use local files,\n"
            error_msg += '    choose "Save to local directory only"\n'
            error_msg += '    in "Input source"\n'

        if (
            self.opt_input.option == "telegram" or self.opt_output.option == "telegram"
        ) and not self.opt_cred.telegram_token:
            error_msg += (
                "[X] Downloading from and uploading to telegram requires bot token.\n"
            )
            error_msg += save_to_local_tip

        if self.opt_output.option == "telegram" and not self.opt_cred.telegram_userid:
            error_msg += "[X] Uploading to telegram requires user_id \n"
            error_msg += "    (From real account, not bot account).\n"
            error_msg += save_to_local_tip

        if self.opt_output.option == "signal" and not (
            self.opt_cred.signal_uuid and self.opt_cred.signal_password
        ):
            error_msg += "[X] Uploading to signal requires uuid and password.\n"
            error_msg += save_to_local_tip

        output_presets = OUTPUT_JSON

        input_option = self.opt_input.option
        output_option = self.opt_output.option

        for metadata in ("title", "author"):
            if MetadataHandler.check_metadata_required(
                output_option, metadata
            ) and not getattr(self.opt_output, metadata):
                if not MetadataHandler.check_metadata_provided(
                    self.opt_input.dir, input_option, metadata
                ):
                    error_msg += f'[X] {output_presets[output_option]["full_name"]} requires {metadata}\n'
                    if self.opt_input.option == "local":
                        error_msg += f"    {metadata} was not supplied and {metadata}.txt is absent\n"
                    else:
                        error_msg += f"    {metadata} was not supplied and input source will not provide {metadata}\n"
                    error_msg += (
                        f"    Supply the {metadata} by filling in the option, or\n"
                    )
                    error_msg += f"    Create {metadata}.txt with the {metadata} name\n"
                else:
                    info_msg += f'[!] {output_presets[output_option]["full_name"]} requires {metadata}\n'
                    if self.opt_input.option == "local":
                        info_msg += f"    {metadata} was not supplied but {metadata}.txt is present\n"
                        info_msg += f"    Using {metadata} name in {metadata}.txt\n"
                    else:
                        info_msg += f"    {metadata} was not supplied but input source will provide {metadata}\n"
                        info_msg += f"    Using {metadata} provided by input source\n"

        if info_msg != "":
            self.executor.cb(info_msg)

        if error_msg != "":
            self.executor.cb(error_msg)
            return False

        # Check if preset not equal to export option
        # Only warn if the compression option is available in export preset
        # Only warn if export option is not local or custom
        # Do not warn if no_compress is true
        if (
            not self.opt_comp.no_compress
            and self.opt_output.option != "local"
            and self.opt_comp.preset != "custom"
            and self.opt_output.option not in self.opt_comp.preset
        ):
            msg = "Compression preset does not match export option\n"
            msg += "You may continue, but the files will need to be compressed again before export\n"
            msg += "You are recommended to choose the matching option for compression and output. Continue?"

            self.executor.cb("ask_bool", (msg,))
            response = self.executor.cb_return.get_response()

            if response is False:
                return False

        for param, value in (
            ("fps_power", self.opt_comp.fps_power),
            ("res_power", self.opt_comp.res_power),
            ("quality_power", self.opt_comp.quality_power),
            ("color_power", self.opt_comp.color_power),
        ):
            if value < -1:
                error_msg += "\n"
                error_msg += f"[X] {param} should be between -1 and positive infinity. {value} was given."

        if self.opt_comp.scale_filter not in (
            "nearest",
            "box",
            "bilinear",
            "hamming",
            "bicubic",
            "lanczos",
        ):
            error_msg += "\n"
            error_msg += (
                f"[X] scale_filter {self.opt_comp.scale_filter} is not valid option"
            )
            error_msg += (
                "    Valid options: nearest, box, bilinear, hamming, bicubic, lanczos"
            )

        if self.opt_comp.quantize_method not in ("imagequant", "fastoctree", "none"):
            error_msg += "\n"
            error_msg += f"[X] quantize_method {self.opt_comp.quantize_method} is not valid option"
            error_msg += "    Valid options: imagequant, fastoctree, none"

        # Warn about unable to download animated Kakao stickers with such link
        if (
            self.opt_output.option == "kakao"
            and urlparse(self.opt_input.url).netloc == "e.kakao.com"
            and not self.opt_cred.kakao_auth_token
        ):
            msg = "To download ANIMATED stickers from e.kakao.com,\n"
            msg += "you need to generate auth_token.\n"
            msg += "Alternatively, you can generate share link (emoticon.kakao.com/items/xxxxx)\n"
            msg += "from Kakao app on phone.\n"
            msg += "You are adviced to read documentations.\n"
            msg += "If you continue, you will only download static stickers. Continue?"

            self.executor.cb("ask_bool", (msg,))
            response = self.executor.cb_return.get_response()

            if response is False:
                return False

        # Warn about in/output directories that might contain other files
        # Directory is safe if the name is stickers_input/stickers_output, or
        # all contents are related to sticker-convert
        for path_type, path, default_name in (
            ("Input", self.opt_input.dir, "stickers_input"),
            ("Output", self.opt_output.dir, "stickers_output"),
        ):
            if path_type == "Input" and (
                path.name == "stickers_input"
                or self.opt_input.option == "local"
                or not any(path.iterdir())
            ):
                continue
            if path_type == "Output" and (
                path.name == "stickers_output"
                or self.opt_comp.no_compress
                or not any(path.iterdir())
            ):
                continue

            related_files = MetadataHandler.get_files_related_to_sticker_convert(path)
            if any(i for i in path.iterdir() if i not in related_files):
                msg = "WARNING: {} directory is set to {}.\n"
                msg += 'It does not have default name of "{}",\n'
                msg += "and It seems like it contains PERSONAL DATA.\n"
                msg += "During execution, contents of this directory\n"
                msg += 'maybe MOVED to "archive_*".\n'
                msg += "THIS MAY CAUSE DAMAGE TO YOUR DATA. Continue?"

                self.executor.cb(
                    "ask_bool", (msg.format(path_type, path, default_name),)
                )
                response = self.executor.cb_return.get_response()

                if response is False:
                    return False

                break

        return True

    def cleanup(self) -> bool:
        # If input is 'From local directory', then we should keep files in input/output directory as it maybe edited by user
        # If input is not 'From local directory', then we should move files in input/output directory as new files will be downloaded
        # Output directory should be cleanup unless no_compress is true (meaning files in output directory might be edited by user)

        timestamp = datetime.now().strftime("%Y-%d-%m_%H-%M-%S")
        dir_name = "archive_" + timestamp

        in_dir_files = MetadataHandler.get_files_related_to_sticker_convert(
            self.opt_input.dir, include_archive=False
        )
        out_dir_files = MetadataHandler.get_files_related_to_sticker_convert(
            self.opt_output.dir, include_archive=False
        )

        if self.opt_input.option == "local":
            self.executor.cb(
                "Skip moving old files in input directory as input source is local"
            )
        elif len(in_dir_files) == 0:
            self.executor.cb(
                "Skip moving old files in input directory as input source is empty"
            )
        else:
            archive_dir = Path(self.opt_input.dir, dir_name)
            self.executor.cb(
                f"Moving old files in input directory to {archive_dir} as input source is not local"
            )
            archive_dir.mkdir(exist_ok=True)
            for old_path in in_dir_files:
                new_path = Path(archive_dir, old_path.name)
                old_path.rename(new_path)

        if self.opt_comp.no_compress:
            self.executor.cb(
                "Skip moving old files in output directory as no_compress is True"
            )
        elif len(out_dir_files) == 0:
            self.executor.cb(
                "Skip moving old files in output directory as output source is empty"
            )
        else:
            archive_dir = Path(self.opt_output.dir, dir_name)
            self.executor.cb(f"Moving old files in output directory to {archive_dir}")
            os.makedirs(archive_dir)
            for old_path in out_dir_files:
                new_path = Path(archive_dir, old_path.name)
                old_path.rename(new_path)

        return True

    def download(self) -> bool:
        downloaders: List[Callable[..., bool]] = []

        if self.opt_input.option == "signal":
            downloaders.append(DownloadSignal.start)

        if self.opt_input.option == "line":
            downloaders.append(DownloadLine.start)

        if self.opt_input.option == "telegram":
            downloaders.append(DownloadTelegram.start)

        if self.opt_input.option == "kakao":
            downloaders.append(DownloadKakao.start)

        if len(downloaders) > 0:
            self.executor.cb("Downloading...")
        else:
            self.executor.cb("Nothing to download")
            return True

        for downloader in downloaders:
            self.executor.add_work(
                work_func=downloader,
                work_args=(self.opt_input.url, self.opt_input.dir, self.opt_cred),
            )

        self.executor.start_workers(processes=1)
        self.executor.join_workers()

        # Return False if any of the job returns failure
        for result in self.executor.get_result():
            if result is False:
                return False

        return True

    def compress(self) -> bool:
        if self.opt_comp.no_compress is True:
            self.executor.cb("no_compress is set to True, skip compression")
            in_dir_files = [
                i
                for i in sorted(self.opt_input.dir.iterdir())
                if Path(self.opt_input.dir, i.name).is_file()
            ]
            out_dir_files = [
                i
                for i in sorted(self.opt_output.dir.iterdir())
                if Path(self.opt_output.dir, i.name).is_file()
            ]
            if len(in_dir_files) == 0:
                self.executor.cb(
                    "Input directory is empty, nothing to copy to output directory"
                )
            elif len(out_dir_files) != 0:
                self.executor.cb(
                    "Output directory is not empty, not copying files from input directory"
                )
            else:
                self.executor.cb(
                    "Output directory is empty, copying files from input directory"
                )
                for i in in_dir_files:
                    src_f = Path(self.opt_input.dir, i.name)
                    dst_f = Path(self.opt_output.dir, i.name)
                    shutil.copy(src_f, dst_f)
            return True
        msg = "Compressing..."

        input_dir = Path(self.opt_input.dir)
        output_dir = Path(self.opt_output.dir)

        in_fs: List[Path] = []

        # .txt: emoji.txt, title.txt
        # .m4a: line sticker sound effects
        for i in sorted(input_dir.iterdir()):
            in_f = input_dir / i

            if not in_f.is_file():
                continue
            if CodecInfo.get_file_ext(i) in (".txt", ".m4a") or Path(i).stem == "cover":
                shutil.copy(in_f, output_dir / i.name)
            else:
                in_fs.append(i)

        in_fs_count = len(in_fs)

        self.executor.cb(msg)
        self.executor.cb(
            "bar", kwargs={"set_progress_mode": "determinate", "steps": in_fs_count}
        )

        for i in in_fs:
            in_f = input_dir / i.name
            out_f = output_dir / Path(i).stem

            self.executor.add_work(
                work_func=StickerConvert.convert, work_args=(in_f, out_f, self.opt_comp)
            )

        self.executor.start_workers(processes=min(self.opt_comp.processes, in_fs_count))
        self.executor.join_workers()

        # Return False if any of the job returns failure
        for result in self.executor.get_result():
            if result[0] is False:
                return False

        return True

    def export(self) -> bool:
        if self.opt_output.option == "local":
            self.executor.cb("Saving to local directory only, nothing to export")
            return True

        self.executor.cb("Exporting...")

        exporters: List[Callable[..., List[str]]] = []

        if self.opt_output.option == "whatsapp":
            exporters.append(CompressWastickers.start)

        if self.opt_output.option == "signal":
            exporters.append(UploadSignal.start)

        if self.opt_output.option == "telegram":
            exporters.append(UploadTelegram.start)

        if self.opt_output.option == "telegram_emoji":
            exporters.append(UploadTelegram.start)

        if self.opt_output.option == "imessage":
            exporters.append(XcodeImessage.start)

        for exporter in exporters:
            self.executor.add_work(
                work_func=exporter,
                work_args=(self.opt_output, self.opt_comp, self.opt_cred),
            )

        self.executor.start_workers(processes=1)
        self.executor.join_workers()

        for result in self.executor.get_result():
            self.out_urls.extend(result)

        if self.out_urls:
            with open(
                Path(self.opt_output.dir, "export-result.txt"), "w+", encoding="utf-8"
            ) as f:
                f.write("\n".join(self.out_urls))
        else:
            self.executor.cb("An error occured while exporting stickers")
            return False

        return True

    def report(self) -> bool:
        msg = "##########\n"
        msg += "Summary:\n"
        msg += "##########\n"
        msg += "\n"

        if self.compress_fails:
            msg += f'Warning: Could not compress the following {len(self.compress_fails)} file{"s" if len(self.compress_fails) > 1 else ""}:\n'
            msg += "\n".join(self.compress_fails)
            msg += "\n"
            msg += "\nConsider adjusting compression parameters"
            msg += "\n"

        if self.out_urls:
            msg += "Export results:\n"
            msg += "\n".join(self.out_urls)
        else:
            msg += "Export result: None"

        self.executor.cb(msg)

        return True



#!/usr/bin/env python3
import os
import platform
from typing import TYPE_CHECKING, Any

from ttkbootstrap import Button, Checkbutton, Label, LabelFrame  # type: ignore

from sticker_convert.definitions import CONFIG_DIR
from sticker_convert.utils.files.run_bin import RunBin

if TYPE_CHECKING:
    from sticker_convert.gui import GUI  # type: ignore


class ConfigFrame(LabelFrame):
    def __init__(self, gui: "GUI", *args: Any, **kwargs: Any) -> None:
        self.gui = gui
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(1, weight=1)
        self.grid_columnconfigure(3, weight=1)

        self.settings_save_cred_lbl = Label(
            self, text="Save credentials", width=18, justify="left", anchor="w"
        )
        self.settings_save_cred_cbox = Checkbutton(
            self,
            variable=self.gui.settings_save_cred_var,
            onvalue=True,
            offvalue=False,
            bootstyle="success-round-toggle",  # type: ignore
        )

        self.settings_clear_cred_lbl = Label(
            self, text="Clear credentials", width=18, justify="left", anchor="w"
        )
        self.settings_clear_cred_btn = Button(
            self,
            text="Clear...",
            command=self.cb_clear_cred,
            bootstyle="secondary",  # type: ignore
        )

        self.settings_restore_default_lbl = Label(
            self, text="Restore default config", width=18, justify="left", anchor="w"
        )
        self.settings_restore_default_btn = Button(
            self,
            text="Restore...",
            command=self.cb_restore_default,
            bootstyle="secondary",  # type: ignore
        )

        self.settings_open_dir_lbl = Label(
            self, text="Config directory", width=18, justify="left", anchor="w"
        )
        self.settings_open_dir_btn = Button(
            self,
            text="Open...",
            command=self.cb_open_config_directory,
            bootstyle="secondary",  # type: ignore
        )

        self.settings_save_cred_lbl.grid(column=0, row=0, sticky="w", padx=3, pady=3)
        self.settings_save_cred_cbox.grid(column=1, row=0, sticky="w", padx=3, pady=3)

        self.settings_clear_cred_lbl.grid(column=2, row=0, sticky="w", padx=3, pady=3)
        self.settings_clear_cred_btn.grid(column=3, row=0, sticky="w", padx=3, pady=3)

        self.settings_open_dir_lbl.grid(column=0, row=1, sticky="w", padx=3, pady=3)
        self.settings_open_dir_btn.grid(column=1, row=1, sticky="w", padx=3, pady=3)

        self.settings_restore_default_lbl.grid(
            column=2, row=1, sticky="w", padx=3, pady=3
        )
        self.settings_restore_default_btn.grid(
            column=3, row=1, sticky="w", padx=3, pady=3
        )

    def cb_clear_cred(self, *_: Any, **kwargs: Any) -> None:
        response = self.gui.cb_ask_bool("Are you sure you want to clear credentials?")
        if response is True:
            self.gui.delete_creds()
            self.gui.load_jsons()
            self.gui.apply_creds()
            self.gui.highlight_fields()
            self.gui.cb_msg_block("Credentials cleared.")

    def cb_restore_default(self, *_: Any, **kwargs: Any) -> None:
        response = self.gui.cb_ask_bool(
            "Are you sure you want to restore default config? (This will not clear credentials.)"
        )
        if response is True:
            self.gui.delete_config()
            self.gui.load_jsons()
            self.gui.apply_config()
            self.gui.highlight_fields()
            self.gui.cb_msg_block("Restored to default config.")

    def cb_open_config_directory(self, *_: Any, **kwargs: Any) -> None:
        self.gui.cb_msg(msg=f"Config is located at {CONFIG_DIR}")
        if platform.system() == "Windows":
            os.startfile(CONFIG_DIR)  # type: ignore
        elif platform.system() == "Darwin":
            RunBin.run_cmd(["open", str(CONFIG_DIR)], silence=True)
        else:
            RunBin.run_cmd(["xdg-open", str(CONFIG_DIR)], silence=True)

    def set_states(self, state: str) -> None:
        self.settings_save_cred_cbox.config(state=state)
        self.settings_clear_cred_btn.config(state=state)
        self.settings_restore_default_btn.config(state=state)
        self.settings_open_dir_btn.config(state=state)



#!/usr/bin/env python3
import argparse
import signal
import sys
from argparse import Namespace
from json.decoder import JSONDecodeError
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Dict

from sticker_convert.definitions import CONFIG_DIR, DEFAULT_DIR
from sticker_convert.job import Job
from sticker_convert.job_option import CompOption, CredOption, InputOption, OutputOption
from sticker_convert.utils.auth.get_kakao_auth import GetKakaoAuth
from sticker_convert.utils.auth.get_line_auth import GetLineAuth
from sticker_convert.utils.auth.get_signal_auth import GetSignalAuth
from sticker_convert.utils.callback import Callback
from sticker_convert.utils.files.json_manager import JsonManager
from sticker_convert.utils.url_detect import UrlDetect
from sticker_convert.version import __version__


class CLI:
    def __init__(self) -> None:
        self.cb = Callback()

    def cli(self) -> None:
        try:
            from sticker_convert.utils.files.json_resources_loader import COMPRESSION_JSON, EMOJI_JSON, HELP_JSON, INPUT_JSON, OUTPUT_JSON
        except RuntimeError as e:
            self.cb.msg(str(e))
            return

        self.help = HELP_JSON
        self.input_presets = INPUT_JSON
        self.compression_presets = COMPRESSION_JSON
        self.output_presets = OUTPUT_JSON
        self.emoji_list = EMOJI_JSON

        parser = argparse.ArgumentParser(
            description="CLI for stickers-convert",
            formatter_class=argparse.RawTextHelpFormatter,
        )

        parser.add_argument("--version", action="version", version=__version__)
        parser.add_argument(
            "--no-confirm",
            dest="no_confirm",
            action="store_true",
            help=self.help["global"]["no_confirm"],
        )

        parser_input = parser.add_argument_group("Input options")
        for k, v_str in self.help["input"].items():
            parser_input.add_argument(f'--{k.replace("_", "-")}', dest=k, help=v_str)
        parser_input_src = parser_input.add_mutually_exclusive_group()
        for k, v_dict in self.input_presets.items():
            if k == "local":
                continue
            parser_input_src.add_argument(
                f'--download-{k.replace("_", "-")}',
                dest=f"download_{k}",
                help=f'{v_dict["help"]}\n({v_dict["example"]})',
            )

        parser_output = parser.add_argument_group("Output options")
        for k, v_str in self.help["output"].items():
            parser_output.add_argument(f'--{k.replace("_", "-")}', dest=k, help=v_str)
        parser_output_dst = parser_output.add_mutually_exclusive_group()
        for k, v_dict in self.output_presets.items():
            if k == "local":
                continue
            parser_output_dst.add_argument(
                f'--export-{k.replace("_", "-")}',
                dest=f"export_{k}",
                action="store_true",
                help=v_dict["help"],
            )

        parser_comp = parser.add_argument_group("Compression options")
        parser_comp.add_argument(
            "--no-compress",
            dest="no_compress",
            action="store_true",
            help=self.help["comp"]["no_compress"],
        )
        parser_comp.add_argument(
            "--preset",
            dest="preset",
            default="auto",
            choices=self.compression_presets.keys(),
            help=self.help["comp"]["preset"],
        )
        flags_comp_int = (
            "steps",
            "processes",
            "fps_min",
            "fps_max",
            "res_min",
            "res_max",
            "res_w_min",
            "res_w_max",
            "res_h_min",
            "res_h_max",
            "quality_min",
            "quality_max",
            "color_min",
            "color_max",
            "duration_min",
            "duration_max",
            "vid_size_max",
            "img_size_max",
        )
        flags_comp_float = ("fps_power", "res_power", "quality_power", "color_power")
        flags_comp_str = (
            "vid_format",
            "img_format",
            "cache_dir",
            "scale_filter",
            "quantize_method",
        )
        flags_comp_bool = ("fake_vid",)
        keyword_args: Dict[str, Any]
        for k, v in self.help["comp"].items():
            if k in flags_comp_int:
                keyword_args = {"type": int, "default": None}
            elif k in flags_comp_float:
                keyword_args = {"type": float, "default": None}
            elif k in flags_comp_str:
                keyword_args = {"default": None}
            elif k in flags_comp_bool:
                keyword_args = {"action": "store_true", "default": None}
            else:
                continue
            parser_comp.add_argument(
                f'--{k.replace("_", "-")}',
                **keyword_args,  # type: ignore
                dest=k,
                help=v,
            )
        parser_comp.add_argument(
            "--default-emoji",
            dest="default_emoji",
            default=self.compression_presets["custom"]["default_emoji"],
            help=self.help["comp"]["default_emoji"],
        )

        parser_cred = parser.add_argument_group("Credentials options")
        flags_cred_bool = ("signal_get_auth", "kakao_get_auth", "line_get_auth")
        for k, v in self.help["cred"].items():
            keyword_args = {}
            if k in flags_cred_bool:
                keyword_args = {"action": "store_true"}
            parser_cred.add_argument(
                f'--{k.replace("_", "-")}',
                **keyword_args,  # type: ignore
                dest=k,
                help=v,
            )

        args = parser.parse_args()

        self.cb.no_confirm = args.no_confirm

        self.opt_input = self.get_opt_input(args)
        self.opt_output = self.get_opt_output(args)
        self.opt_comp = self.get_opt_comp(args)
        self.opt_cred = self.get_opt_cred(args)

        job = Job(
            self.opt_input,
            self.opt_comp,
            self.opt_output,
            self.opt_cred,
            self.cb.msg,
            self.cb.msg_block,
            self.cb.bar,
            self.cb.ask_bool,
            self.cb.ask_str,
        )

        signal.signal(signal.SIGINT, job.cancel)
        status = job.start()
        sys.exit(status)

    def get_opt_input(self, args: Namespace) -> InputOption:
        download_options = {
            "auto": args.download_auto,
            "signal": args.download_signal,
            "line": args.download_line,
            "telegram": args.download_telegram,
            "kakao": args.download_kakao,
        }

        download_option = "local"
        url = ""
        for k, v in download_options.items():
            if v:
                download_option = k
                url = v
                break

        if download_option == "auto":
            detected_download_option = UrlDetect.detect(url)
            if detected_download_option:
                download_option = detected_download_option
                self.cb.msg(f"Detected URL input source: {download_option}")
            else:
                self.cb.msg(f"Error: Unrecognied URL input source for url: {url}")
                sys.exit()

        opt_input = InputOption(
            option=download_option,
            url=url,
            dir=Path(args.input_dir).resolve()
            if args.input_dir
            else DEFAULT_DIR / "stickers_input",
        )

        return opt_input

    def get_opt_output(self, args: Namespace) -> OutputOption:
        if args.export_whatsapp:
            export_option = "whatsapp"
        elif args.export_signal:
            export_option = "signal"
        elif args.export_telegram:
            export_option = "telegram"
        elif args.export_imessage:
            export_option = "imessage"
        else:
            export_option = "local"

        opt_output = OutputOption(
            option=export_option,
            dir=Path(args.output_dir).resolve()
            if args.output_dir
            else DEFAULT_DIR / "stickers_output",
            title=args.title,
            author=args.author,
        )

        return opt_output

    def get_opt_comp(self, args: Namespace) -> CompOption:
        preset: str = args.preset if args.preset else "auto"
        if args.preset == "custom":
            if (
                sum(
                    (
                        args.export_whatsapp,
                        args.export_signal,
                        args.export_telegram,
                        args.export_imessage,
                    )
                )
                > 1
            ):
                # Let the verify functions in export do the compression
                args.no_compress = True
            elif args.export_whatsapp:
                preset = "whatsapp"
            elif args.export_signal:
                preset = "signal"
            elif args.export_telegram:
                preset = "telegram"
            elif args.export_imessage:
                preset = "imessage_small"
        elif args.preset == "auto":
            output_option = (
                self.opt_output.option if self.opt_output.option else "local"
            )
            if output_option == "local":
                preset = "custom"
                args.no_compress = True
                self.cb.msg(
                    "Auto compression option set to no_compress (Reason: Export to local directory only)"
                )
            elif output_option == "imessage":
                preset = "imessage_small"
                self.cb.msg(f"Auto compression option set to {preset}")
            else:
                preset = output_option
                self.cb.msg(f"Auto compression option set to {preset}")

        opt_comp = CompOption(
            preset=preset,
            size_max_img=self.compression_presets[preset]["size_max"]["img"]
            if args.img_size_max is None
            else args.img_size_max,
            size_max_vid=self.compression_presets[preset]["size_max"]["vid"]
            if args.vid_size_max is None
            else args.vid_size_max,
            format_img=(
                self.compression_presets[preset]["format"]["img"]
                if args.img_format is None
                else args.img_format,
            ),
            format_vid=(
                self.compression_presets[preset]["format"]["vid"]
                if args.vid_format is None
                else args.vid_format,
            ),
            fps_min=self.compression_presets[preset]["fps"]["min"]
            if args.fps_min is None
            else args.fps_min,
            fps_max=self.compression_presets[preset]["fps"]["max"]
            if args.fps_max is None
            else args.fps_max,
            fps_power=self.compression_presets[preset]["fps"]["power"]
            if args.fps_power is None
            else args.fps_power,
            res_w_min=self.compression_presets[preset]["res"]["w"]["min"]
            if args.res_w_min is None
            else args.res_w_min,
            res_w_max=self.compression_presets[preset]["res"]["w"]["max"]
            if args.res_w_max is None
            else args.res_w_max,
            res_h_min=self.compression_presets[preset]["res"]["h"]["min"]
            if args.res_h_min is None
            else args.res_h_min,
            res_h_max=self.compression_presets[preset]["res"]["h"]["max"]
            if args.res_h_max is None
            else args.res_h_max,
            res_power=self.compression_presets[preset]["res"]["power"]
            if args.res_power is None
            else args.res_power,
            quality_min=self.compression_presets[preset]["quality"]["min"]
            if args.quality_min is None
            else args.quality_min,
            quality_max=self.compression_presets[preset]["quality"]["max"]
            if args.quality_max is None
            else args.quality_max,
            quality_power=self.compression_presets[preset]["quality"]["power"]
            if args.quality_power is None
            else args.quality_power,
            color_min=self.compression_presets[preset]["color"]["min"]
            if args.color_min is None
            else args.color_min,
            color_max=self.compression_presets[preset]["color"]["max"]
            if args.color_max is None
            else args.color_max,
            color_power=self.compression_presets[preset]["color"]["power"]
            if args.color_power is None
            else args.color_power,
            duration_min=self.compression_presets[preset]["duration"]["min"]
            if args.duration_min is None
            else args.duration_min,
            duration_max=self.compression_presets[preset]["duration"]["max"]
            if args.duration_max is None
            else args.duration_max,
            steps=self.compression_presets[preset]["steps"]
            if args.steps is None
            else args.steps,
            fake_vid=self.compression_presets[preset]["fake_vid"]
            if args.fake_vid is None
            else args.fake_vid,
            cache_dir=args.cache_dir,
            scale_filter=self.compression_presets[preset]["scale_filter"]
            if args.scale_filter is None
            else args.scale_filter,
            quantize_method=self.compression_presets[preset]["quantize_method"]
            if args.quantize_method is None
            else args.quantize_method,
            default_emoji=self.compression_presets[preset]["default_emoji"]
            if args.default_emoji is None
            else args.default_emoji,
            no_compress=args.no_compress,
            processes=args.processes if args.processes else ceil(cpu_count() / 2),
        )

        return opt_comp

    def get_opt_cred(self, args: Namespace) -> CredOption:
        creds_path = CONFIG_DIR / "creds.json"
        creds = {}
        if creds_path.is_file():
            try:
                creds = JsonManager.load_json(creds_path)
            except JSONDecodeError:
                self.cb.msg("Warning: creds.json content is corrupted")
                creds = {}
        else:
            creds = {}

        if creds:
            self.cb.msg("Loaded credentials from creds.json")

        opt_cred = CredOption(
            signal_uuid=args.signal_uuid
            if args.signal_uuid
            else creds.get("signal", {}).get("uuid"),
            signal_password=args.signal_password
            if args.signal_password
            else creds.get("signal", {}).get("password"),
            telegram_token=args.telegram_token
            if args.telegram_token
            else creds.get("telegram", {}).get("token"),
            telegram_userid=args.telegram_userid
            if args.telegram_userid
            else creds.get("telegram", {}).get("userid"),
            kakao_auth_token=args.kakao_auth_token
            if args.kakao_auth_token
            else creds.get("kakao", {}).get("auth_token"),
            kakao_username=args.kakao_username
            if args.kakao_username
            else creds.get("kakao", {}).get("username"),
            kakao_password=args.kakao_password
            if args.kakao_password
            else creds.get("kakao", {}).get("password"),
            kakao_country_code=args.kakao_country_code
            if args.kakao_country_code
            else creds.get("kakao", {}).get("country_code"),
            kakao_phone_number=args.kakao_phone_number
            if args.kakao_phone_number
            else creds.get("kakao", {}).get("phone_number"),
            line_cookies=args.line_cookies
            if args.line_cookies
            else creds.get("line", {}).get("cookies"),
        )

        if args.kakao_get_auth:
            get_kakao_auth = GetKakaoAuth(
                opt_cred=opt_cred,
                cb_msg=self.cb.msg,
                cb_msg_block=self.cb.msg_block,
                cb_ask_str=self.cb.ask_str,
            )
            auth_token = get_kakao_auth.get_cred()

            if auth_token:
                opt_cred.kakao_auth_token = auth_token

                self.cb.msg(f"Got auth_token successfully: {auth_token}")

        if args.signal_get_auth:
            get_signal_auth = GetSignalAuth()

            signal_bin_path = None
            signal_user_data_dir = None
            if args.signal_data_dir:
                signal_bin_path = "(User specified)"
                signal_user_data_dir = args.signal_data_dir

            uuid, password, msg = get_signal_auth.get_cred(
                signal_bin_path, signal_user_data_dir
            )

            if uuid and password:
                opt_cred.signal_uuid = uuid
                opt_cred.signal_password = password

            self.cb.msg(msg)

        if args.line_get_auth:
            get_line_auth = GetLineAuth()

            line_cookies = get_line_auth.get_cred()

            if line_cookies:
                opt_cred.line_cookies = line_cookies

                self.cb.msg("Got Line cookies successfully")
            else:
                self.cb.msg(
                    "Failed to get Line cookies. Have you logged in the web browser?"
                )

        if args.save_cred:
            creds_path = CONFIG_DIR / "creds.json"
            JsonManager.save_json(creds_path, opt_cred.to_dict())
            self.cb.msg("Saved credentials to creds.json")

        return opt_cred



#!/usr/bin/env python3
import copy
import json
import os
import plistlib
import shutil
import zipfile
from pathlib import Path
from queue import Queue
from typing import Any, Dict, List, Union

from sticker_convert.converter import StickerConvert
from sticker_convert.definitions import ROOT_DIR
from sticker_convert.job_option import CompOption, CredOption, OutputOption
from sticker_convert.uploaders.upload_base import UploadBase
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.metadata_handler import XCODE_IMESSAGE_ICONSET, MetadataHandler
from sticker_convert.utils.files.sanitize_filename import sanitize_filename
from sticker_convert.utils.media.codec_info import CodecInfo
from sticker_convert.utils.media.format_verify import FormatVerify


class XcodeImessage(UploadBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.base_spec.set_size_max(500000)
        self.base_spec.set_res(300)
        self.base_spec.set_format(("png", ".apng", ".gif", ".jpeg", "jpg"))
        self.base_spec.square = True

        self.small_spec = copy.deepcopy(self.base_spec)

        self.medium_spec = copy.deepcopy(self.base_spec)
        self.medium_spec.set_res(408)

        self.large_spec = copy.deepcopy(self.base_spec)
        self.large_spec.set_res(618)

    def create_imessage_xcode(self) -> List[str]:
        urls: List[str] = []
        title, author, _ = MetadataHandler.get_metadata(
            self.opt_output.dir,
            title=self.opt_output.title,
            author=self.opt_output.author,
        )
        if not author:
            self.cb.put("author is required for creating Xcode iMessage sticker pack")
            return urls
        if not title:
            self.cb.put("title is required for creating Xcode iMessage sticker pack")
            return urls

        author = author.replace(" ", "_")
        title = title.replace(" ", "_")
        packs = MetadataHandler.split_sticker_packs(
            self.opt_output.dir,
            title=title,
            file_per_pack=100,
            separate_image_anim=False,
        )

        res_choice = None
        spec_choice = None
        opt_comp_merged = copy.deepcopy(self.opt_comp)

        for pack_title, stickers in packs.items():
            pack_title = sanitize_filename(pack_title)

            for src in stickers:
                self.cb.put(f"Verifying {src} for creating Xcode iMessage sticker pack")

                fpath = Path(self.opt_output.dir, src)

                if res_choice is None:
                    res_choice, _ = CodecInfo.get_file_res(fpath)

                    if res_choice == 618:
                        spec_choice = self.large_spec
                    elif res_choice == 408:
                        spec_choice = self.medium_spec
                    else:
                        # res_choice == 300
                        spec_choice = self.small_spec

                    opt_comp_merged.merge(spec_choice)

                assert spec_choice
                if not FormatVerify.check_file(src, spec=spec_choice):
                    StickerConvert.convert(
                        fpath, fpath, opt_comp_merged, self.cb, self.cb_return
                    )

            self.add_metadata(author, pack_title)
            self.create_xcode_proj(author, pack_title)

            result = Path(self.opt_output.dir / pack_title).as_posix()
            self.cb.put(result)
            urls.append(result)

        return urls

    def add_metadata(self, author: str, title: str) -> None:
        first_image_path = Path(
            self.opt_output.dir,
            [
                i
                for i in sorted(self.opt_output.dir.iterdir())
                if (self.opt_output.dir / i).is_file() and i.suffix == ".png"
            ][0],
        )
        cover_path = MetadataHandler.get_cover(self.opt_output.dir)
        if cover_path:
            icon_source = cover_path
        else:
            icon_source = first_image_path

        for icon, res in XCODE_IMESSAGE_ICONSET.items():
            spec_cover = CompOption()
            spec_cover.set_res_w(res[0])
            spec_cover.set_res_h(res[1])
            spec_cover.set_fps(0)

            icon_path = self.opt_output.dir / icon
            if Path(icon) in list(
                self.opt_output.dir.iterdir()
            ) and not FormatVerify.check_file(icon_path, spec=spec_cover):
                StickerConvert.convert(
                    icon_path, icon_path, spec_cover, self.cb, self.cb_return
                )
            else:
                StickerConvert.convert(
                    icon_source, icon_path, spec_cover, self.cb, self.cb_return
                )

        MetadataHandler.set_metadata(self.opt_output.dir, author=author, title=title)

    def create_xcode_proj(self, author: str, title: str) -> None:
        pack_path = self.opt_output.dir / title
        if (ROOT_DIR / "ios-message-stickers-template.zip").is_file():
            with zipfile.ZipFile(
                ROOT_DIR / "ios-message-stickers-template.zip", "r"
            ) as f:
                f.extractall(pack_path)
        elif (ROOT_DIR / "ios-message-stickers-template").is_dir():
            shutil.copytree(ROOT_DIR / "ios-message-stickers-template", pack_path)
        else:
            self.cb.put(
                "Failed to create Xcode project: ios-message-stickers-template not found"
            )

        os.remove(pack_path / "README.md")
        shutil.rmtree(
            pack_path / "stickers.xcodeproj/project.xcworkspace",
            ignore_errors=True,
        )
        shutil.rmtree(pack_path / "stickers.xcodeproj/xcuserdata", ignore_errors=True)

        with open(
            pack_path / "stickers.xcodeproj/project.pbxproj",
            encoding="utf-8",
        ) as f:
            pbxproj_data = f.read()

        pbxproj_data = pbxproj_data.replace(
            "stickers StickerPackExtension", f"{title} StickerPackExtension"
        )
        pbxproj_data = pbxproj_data.replace("stickers.app", f"{title}.app")
        pbxproj_data = pbxproj_data.replace("/* stickers */", f"/* {title} */")
        pbxproj_data = pbxproj_data.replace("name = stickers", f"name = {title}")
        pbxproj_data = pbxproj_data.replace(
            "productName = stickers", f"productName = {title}"
        )
        pbxproj_data = pbxproj_data.replace(
            '/* Build configuration list for PBXProject "stickers" */',
            f'/* Build configuration list for PBXProject "{title}" */',
        )
        pbxproj_data = pbxproj_data.replace(
            '/* Build configuration list for PBXNativeTarget "stickers StickerPackExtension" */',
            f'/* Build configuration list for PBXNativeTarget "{title} StickerPackExtension" */',
        )
        pbxproj_data = pbxproj_data.replace(
            '/* Build configuration list for PBXNativeTarget "stickers" */',
            f'/* Build configuration list for PBXNativeTarget "{title}" */',
        )
        pbxproj_data = pbxproj_data.replace("com.niklaspeterson", f"com.{author}")
        pbxproj_data = pbxproj_data.replace(
            "stickers/Info.plist", f"{title}/Info.plist"
        )

        with open(
            pack_path / "stickers.xcodeproj/project.pbxproj",
            "w+",
            encoding="utf-8",
        ) as f:
            f.write(pbxproj_data)

        # packname StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack
        stickers_path = (
            pack_path
            / "stickers StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack"
        )

        for i in stickers_path.iterdir():
            if i.suffix == ".sticker":
                shutil.rmtree(stickers_path / i)

        stickers_lst: List[str] = []
        for i in sorted(self.opt_output.dir.iterdir()):
            if (
                CodecInfo.get_file_ext(i) == ".png"
                and i.stem != "cover"
                and i.name not in XCODE_IMESSAGE_ICONSET
            ):
                sticker_dir = f"{i.stem}.sticker"  # 0.sticker
                stickers_lst.append(sticker_dir)
                # packname StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack/0.sticker
                sticker_path = stickers_path / sticker_dir
                os.mkdir(sticker_path)
                # packname StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack/0.sticker/0.png
                shutil.copy(self.opt_output.dir / i.name, sticker_path / i.name)

                sticker_json_content = {
                    "info": {
                        "author": "xcode",
                        "version": 1,
                    },
                    "properties": {"filename": str(i)},
                }

                # packname StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack/0.sticker/Contents.json
                with open(sticker_path / "Contents.json", "w+", encoding="utf-8") as f:
                    json.dump(sticker_json_content, f, indent=2)

        # packname StickerPackExtension/Stickers.xcstickers/Sticker Pack.stickerpack/Contents.json
        with open(stickers_path / "Contents.json", encoding="utf-8") as f:
            stickerpack_json_content: Dict[str, List[Dict[str, str]]] = json.load(f)

        stickerpack_json_content["stickers"] = []
        for sticker in stickers_lst:
            stickerpack_json_content["stickers"].append({"filename": sticker})

        with open(stickers_path / "Contents.json", "w+", encoding="utf-8") as f:
            json.dump(stickerpack_json_content, f, indent=2)

        # packname StickerPackExtension/Stickers.xcstickers/iMessage App Icon.stickersiconset
        iconset_path = (
            pack_path
            / "stickers StickerPackExtension/Stickers.xcstickers/iMessage App Icon.stickersiconset"
        )

        for iconfile_name in iconset_path.iterdir():
            if Path(iconfile_name).suffix == ".png":
                os.remove(iconset_path / iconfile_name)

        icons_lst: List[str] = []
        for icon in XCODE_IMESSAGE_ICONSET:
            shutil.copy(self.opt_output.dir / icon, iconset_path / icon)
            icons_lst.append(icon)

        # packname/Info.plist
        plist_path = pack_path / "stickers/Info.plist"
        with open(plist_path, "rb") as f:
            plist_dict = plistlib.load(f)
        plist_dict["CFBundleDisplayName"] = title

        with open(plist_path, "wb+") as f:
            plistlib.dump(plist_dict, f)

        Path(pack_path, "stickers").rename(Path(pack_path, title))
        Path(pack_path, "stickers StickerPackExtension").rename(
            Path(pack_path, f"{title} StickerPackExtension")
        )
        Path(pack_path, "stickers.xcodeproj").rename(
            Path(pack_path, f"{title}.xcodeproj")
        )

    @staticmethod
    def start(
        opt_output: OutputOption,
        opt_comp: CompOption,
        opt_cred: CredOption,
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> List[str]:
        exporter = XcodeImessage(opt_output, opt_comp, opt_cred, cb, cb_return)
        return exporter.create_imessage_xcode()



#!/usr/bin/env python3
from __future__ import annotations

import json
from dataclasses import dataclass
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union


def to_int(i: Union[float, str, None]) -> Optional[int]:
    return int(i) if i is not None else None


@dataclass
class BaseOption:
    def merge(self, config: "BaseOption") -> None:
        for k, v in vars(config).items():
            if v is not None:
                setattr(self, k, v)

    def __repr__(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self) -> Dict[str, str]:
        return {}


@dataclass
class InputOption(BaseOption):
    option: str = "local"
    url: str = ""
    dir: Path = Path()

    def to_dict(self) -> Dict[Any, Any]:
        return {"option": self.option, "url": self.url, "dir": self.dir.as_posix()}


@dataclass
class CompOption(BaseOption):
    preset: str = "auto"
    size_max_img: Optional[int] = None
    size_max_vid: Optional[int] = None

    format_img: Tuple[str, ...] = tuple()
    format_vid: Tuple[str, ...] = tuple()

    fps_min: Optional[int] = None
    fps_max: Optional[int] = None
    fps_power: float = -0.5

    res_w_min: Optional[int] = None
    res_w_max: Optional[int] = None
    res_h_min: Optional[int] = None
    res_h_max: Optional[int] = None
    res_power: float = 3.0

    quality_min: Optional[int] = None
    quality_max: Optional[int] = None
    quality_power: float = 5.0

    color_min: Optional[int] = None
    color_max: Optional[int] = None
    color_power: float = 3.0

    duration_min: Optional[int] = None
    duration_max: Optional[int] = None

    steps: int = 1
    fake_vid: Optional[bool] = None
    quantize_method: Optional[str] = None
    scale_filter: Optional[str] = None
    cache_dir: Optional[str] = None
    default_emoji: str = "😀"
    no_compress: Optional[bool] = None
    processes: int = ceil(cpu_count() / 2)
    animated: Optional[bool] = None
    square: Optional[bool] = None

    def to_dict(self) -> Dict[Any, Any]:
        return {
            "preset": self.preset,
            "size_max": {"img": self.size_max_img, "vid": self.size_max_vid},
            "format": {"img": self.format_img, "vid": self.format_vid},
            "fps": {"min": self.fps_min, "max": self.fps_max, "power": self.fps_power},
            "res": {
                "w": {"min": self.res_w_min, "max": self.res_w_max},
                "h": {"min": self.res_h_min, "max": self.res_h_max},
                "power": self.res_power,
            },
            "quality": {
                "min": self.quality_min,
                "max": self.quality_max,
                "power": self.quality_power,
            },
            "color": {
                "min": self.color_min,
                "max": self.color_max,
                "power": self.color_power,
            },
            "duration": {"min": self.duration_min, "max": self.duration_max},
            "steps": self.steps,
            "fake_vid": self.fake_vid,
            "scale_filter": self.scale_filter,
            "cache_dir": self.cache_dir,
            "default_emoji": self.default_emoji,
            "no_compress": self.no_compress,
            "processes": self.processes,
            "animated": self.animated,
            "square": self.square,
        }

    def get_size_max(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.size_max_img, self.size_max_vid)

    def set_size_max(self, value: Optional[int]) -> None:
        self.size_max_img, self.size_max_vid = to_int(value), to_int(value)

    def get_format(self) -> Tuple[Tuple[str, ...], Tuple[str, ...]]:
        return (self.format_img, self.format_vid)

    def set_format(self, value: Tuple[str, ...]) -> None:
        self.format_img, self.format_vid = value, value

    def get_fps(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.fps_min, self.fps_max)

    def set_fps(self, value: Optional[int]) -> None:
        self.fps_min, self.fps_max = to_int(value), to_int(value)

    def get_res(
        self,
    ) -> Tuple[
        Tuple[Optional[int], Optional[int]], Tuple[Optional[int], Optional[int]]
    ]:
        return (self.get_res_w(), self.get_res_h())

    def set_res(self, value: Optional[int]) -> None:
        self.res_w_min = to_int(value)
        self.res_w_max = to_int(value)
        self.res_h_min = to_int(value)
        self.res_h_max = to_int(value)

    def get_res_max(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.res_w_max, self.res_h_max)

    def set_res_max(self, value: Optional[int]) -> None:
        self.res_w_max = to_int(value)
        self.res_h_max = to_int(value)

    def get_res_min(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.res_w_min, self.res_h_min)

    def set_res_min(self, value: Optional[int]) -> None:
        self.res_w_min = to_int(value)
        self.res_h_min = to_int(value)

    def get_res_w(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.res_w_min, self.res_w_max)

    def set_res_w(self, value: Optional[int]) -> None:
        self.res_w_min, self.res_w_max = to_int(value), to_int(value)

    def get_res_h(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.res_h_min, self.res_h_max)

    def set_res_h(self, value: Optional[int]) -> None:
        self.res_h_min, self.res_h_max = to_int(value), to_int(value)

    def get_quality(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.quality_min, self.quality_max)

    def set_quality(self, value: Optional[int]) -> None:
        self.quality_min, self.quality_max = to_int(value), to_int(value)

    def get_color(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.color_min, self.color_max)

    def set_color(self, value: Optional[int]) -> None:
        self.color_min, self.color_max = to_int(value), to_int(value)

    def get_duration(self) -> Tuple[Optional[int], Optional[int]]:
        return (self.duration_min, self.duration_max)

    def set_duration(self, value: Optional[int]) -> None:
        self.duration_min, self.duration_max = to_int(value), to_int(value)


@dataclass
class OutputOption(BaseOption):
    option: str = "local"
    dir: Path = Path()
    title: str = ""
    author: str = ""

    def to_dict(self) -> Dict[Any, Any]:
        return {
            "option": self.option,
            "dir": self.dir.as_posix(),
            "title": self.title,
            "author": self.author,
        }


@dataclass
class CredOption(BaseOption):
    signal_uuid: str = ""
    signal_password: str = ""
    telegram_token: str = ""
    telegram_userid: str = ""
    kakao_auth_token: str = ""
    kakao_username: str = ""
    kakao_password: str = ""
    kakao_country_code: str = ""
    kakao_phone_number: str = ""
    line_cookies: str = ""

    def to_dict(self) -> Dict[Any, Any]:
        return {
            "signal": {"uuid": self.signal_uuid, "password": self.signal_password},
            "telegram": {"token": self.telegram_token, "userid": self.telegram_userid},
            "kakao": {
                "auth_token": self.kakao_auth_token,
                "username": self.kakao_username,
                "password": self.kakao_password,
                "country_code": self.kakao_country_code,
                "phone_number": self.kakao_phone_number,
            },
            "line": {"cookies": self.line_cookies},
        }



#!/usr/bin/env python3
from typing import Any

from ttkbootstrap import Menu  # type: ignore


# Reference: https://stackoverflow.com/a/57704013
class RightClicker:
    def __init__(self, event: Any) -> None:
        right_click_menu = Menu(None, tearoff=0, takefocus=0)

        for txt in ["Cut", "Copy", "Paste"]:
            right_click_menu.add_command(
                label=txt,
                command=lambda event=event, text=txt: self.right_click_command(
                    event, text
                ),
            )

        right_click_menu.tk_popup(event.x_root, event.y_root, entry="0")

    def right_click_command(self, event: Any, cmd: str) -> None:
        event.widget.event_generate(f"<<{cmd}>>")



#!/usr/bin/env python3
from multiprocessing import Event, Queue
from typing import Any, Callable, Dict, Optional, Tuple, Union

from tqdm import tqdm

CbQueueTupleType = Tuple[
    Optional[str], Optional[Tuple[Any, ...]], Optional[Dict[str, Any]]
]
CbQueueItemType = Union[
    CbQueueTupleType,
    str,
    None,
]


class CallbackReturn:
    def __init__(self) -> None:
        self.response_event = Event()
        self.response_queue: Queue[Union[bool, str, None]] = Queue()

    def set_response(self, response: Union[bool, str, None]) -> None:
        self.response_queue.put(response)
        self.response_event.set()

    def get_response(self) -> Union[bool, str, None]:
        self.response_event.wait()

        response = self.response_queue.get()

        self.response_event.clear()

        return response


class Callback:
    def __init__(
        self,
        msg: Optional[Callable[..., None]] = None,
        bar: Optional[Callable[..., None]] = None,
        msg_block: Optional[Callable[..., None]] = None,
        ask_bool: Optional[Callable[..., bool]] = None,
        ask_str: Optional[Callable[..., str]] = None,
        silent: bool = False,
        no_confirm: bool = False,
    ) -> None:
        self.progress_bar: Optional[tqdm[Any]] = None

        if msg:
            self.msg = msg
        else:
            self.msg = self.cb_msg

        if bar:
            self.bar = bar
        else:
            self.bar = self.cb_bar

        if msg_block:
            self.msg_block = msg_block
        else:
            self.msg_block = self.cb_msg_block

        if ask_bool:
            self.ask_bool = ask_bool
        else:
            self.ask_bool = self.cb_ask_bool

        if ask_str:
            self.ask_str = ask_str
        else:
            self.ask_str = self.cb_ask_str

        self.silent = silent
        self.no_confirm = no_confirm

    def cb_msg(self, *args: Any, **kwargs: Any) -> None:
        if self.silent:
            return

        msg = kwargs.get("msg")
        file = kwargs.get("file")

        if not msg and len(args) == 1:
            msg = str(args[0])

        if msg:
            if self.progress_bar:
                self.progress_bar.write(msg)
            elif file:
                print(msg, file=file)
            else:
                print(msg)

    def cb_bar(
        self,
        set_progress_mode: Optional[str] = None,
        steps: int = 0,
        update_bar: bool = False,
    ) -> None:
        if self.silent:
            return

        if self.progress_bar:
            if update_bar:
                self.progress_bar.update()
            elif set_progress_mode == "indeterminate":
                self.progress_bar.close()
                self.progress_bar = None
            elif set_progress_mode == "clear":
                self.progress_bar.reset()
        elif set_progress_mode == "determinate":
            self.progress_bar = tqdm(total=steps)

    def cb_msg_block(self, *args: Any) -> None:
        if self.silent:
            return
        if len(args) > 0:
            msg = " ".join(str(i) for i in args)
            self.msg(msg)
        if not self.no_confirm:
            input("Press Enter to continue...")

    def cb_ask_bool(self, *args: Any, **kwargs: Any) -> bool:
        question = args[0]

        self.msg(question)

        if self.no_confirm:
            self.msg(
                '"--no-confirm" flag is set. Continue with this run without asking questions'
            )
            return True

        self.msg(
            '[If you do not want to get asked by this question, add "--no-confirm" flag]'
        )
        self.msg()
        result = input("Continue? [y/N] > ")
        if result.lower() != "y":
            return False
        return True

    def cb_ask_str(
        self,
        msg: Optional[str] = None,
        initialvalue: Optional[str] = None,
        cli_show_initialvalue: bool = True,
    ) -> str:
        self.msg(msg)

        hint = ""
        if cli_show_initialvalue and initialvalue:
            hint = f" [Default: {initialvalue}]"

        response = input(f"Enter your response and press enter{hint} > ")

        if initialvalue and not response:
            response = initialvalue

        return response

    def put(
        self,
        i: Union[
            CbQueueItemType,
            str,
        ],
    ) -> Union[str, bool, None]:
        if isinstance(i, tuple):
            action = i[0]
            if len(i) >= 2:
                args: Tuple[str, ...] = i[1] if i[1] else tuple()
            else:
                args = tuple()
            if len(i) >= 3:
                kwargs: Dict[str, Any] = i[2] if i[2] else {}
            else:
                kwargs = {}
        else:
            action = i
            args = tuple()
            kwargs = {}

        # Fake implementation for Queue.put()
        if action is None:
            return None
        if action == "msg":
            self.msg(*args, **kwargs)
        elif action == "bar":
            self.bar(**kwargs)
        elif action == "update_bar":
            self.bar(update_bar=True)
        elif action == "msg_block":
            return self.msg_block(*args, **kwargs)
        elif action == "ask_bool":
            return self.ask_bool(*args, **kwargs)
        elif action == "ask_str":
            return self.ask_str(**kwargs)
        else:
            self.msg(action)
        return None



#!/usr/bin/env python3
from pathlib import Path
from tkinter import filedialog
from typing import TYPE_CHECKING, Any

from ttkbootstrap import Button, Entry, Label, LabelFrame, OptionMenu  # type: ignore

from sticker_convert.definitions import DEFAULT_DIR
from sticker_convert.gui_components.frames.right_clicker import RightClicker

if TYPE_CHECKING:
    from sticker_convert.gui import GUI  # type: ignore


class OutputFrame(LabelFrame):
    def __init__(self, gui: "GUI", *args: Any, **kwargs: Any) -> None:
        self.gui = gui
        super().__init__(*args, **kwargs)

        self.output_option_lbl = Label(
            self, text="Output options", width=18, justify="left", anchor="w"
        )
        output_full_names = [i["full_name"] for i in self.gui.output_presets.values()]
        defult_output_full_name = self.gui.output_presets[self.gui.default_output_mode][
            "full_name"
        ]
        self.output_option_opt = OptionMenu(
            self,
            self.gui.output_option_display_var,
            defult_output_full_name,
            *output_full_names,
            command=self.cb_output_option,
            bootstyle="secondary",  # type: ignore
        )
        self.output_option_opt.config(width=32)

        self.output_setdir_lbl = Label(
            self, text="Output directory", justify="left", anchor="w"
        )
        self.output_setdir_entry = Entry(
            self,
            textvariable=self.gui.output_setdir_var,
            width=60,
            validatecommand=self.gui.highlight_fields,
        )
        self.output_setdir_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.output_setdir_btn = Button(
            self,
            text="Choose directory",
            command=self.cb_set_outdir,
            width=16,
            bootstyle="secondary",  # type: ignore
        )

        self.title_lbl = Label(self, text="Title")
        self.title_entry = Entry(
            self,
            textvariable=self.gui.title_var,
            width=80,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.title_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.author_lbl = Label(self, text="Author")
        self.author_entry = Entry(
            self,
            textvariable=self.gui.author_var,
            width=80,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.author_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.output_option_lbl.grid(column=0, row=0, sticky="w", padx=3, pady=3)
        self.output_option_opt.grid(
            column=1, row=0, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.output_setdir_lbl.grid(
            column=0, row=1, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.output_setdir_entry.grid(column=1, row=1, sticky="w", padx=3, pady=3)
        self.output_setdir_btn.grid(column=2, row=1, sticky="e", padx=3, pady=3)
        self.title_lbl.grid(column=0, row=2, sticky="w", padx=3, pady=3)
        self.title_entry.grid(column=1, columnspan=2, row=2, sticky="w", padx=3, pady=3)
        self.author_lbl.grid(column=0, row=3, sticky="w", padx=3, pady=3)
        self.author_entry.grid(
            column=1, columnspan=2, row=3, sticky="w", padx=3, pady=3
        )

    def cb_set_outdir(self, *_args: Any) -> None:
        orig_output_dir = self.gui.output_setdir_var.get()
        if not Path(orig_output_dir).is_dir():
            orig_output_dir = DEFAULT_DIR
        output_dir = filedialog.askdirectory(initialdir=orig_output_dir)
        if output_dir:
            self.gui.output_setdir_var.set(output_dir)

    def cb_output_option(self, *_: Any) -> None:
        self.gui.output_option_true_var.set(self.gui.output_option_display_var.get())
        self.gui.comp_frame.cb_comp_apply_preset()
        self.gui.highlight_fields()

    def set_states(self, state: str) -> None:
        self.title_entry.config(state=state)
        self.author_entry.config(state=state)
        self.output_option_opt.config(state=state)
        self.output_setdir_entry.config(state=state)
        self.output_setdir_btn.config(state=state)



#!/usr/bin/env python3
import re
import unicodedata

BLACKLIST_CHAR = ("\\", "/", ":", "*", "?", '"', "<", ">", "|", "\0")
RESERVED_FILENAME = (
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
)  # Reserved words on Windows


def sanitize_filename(filename: str) -> str:
    # Based on https://gitlab.com/jplusplus/sanitize-filename/-/blob/master/sanitize_filename/sanitize_filename.py
    # Replace illegal character with '_'
    """Return a fairly safe version of the filename.

    We don't limit ourselves to ascii, because we want to keep municipality
    names, etc, but we do want to get rid of anything potentially harmful,
    and make sure we do not exceed Windows filename length limits.
    Hence a less safe blacklist, rather than a whitelist.
    """

    filename = "".join(c if c not in BLACKLIST_CHAR else "_" for c in filename)
    # Remove all charcters below code point 32
    filename = "".join(c if 31 < ord(c) else "_" for c in filename)
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.rstrip(". ")  # Windows does not allow these at end
    filename = filename.strip()
    if all(x == "." for x in filename):
        filename = "__" + filename
    if filename in RESERVED_FILENAME:
        filename = "__" + filename
    if len(filename) == 0:
        filename = "__"
    if len(filename) > 255:
        parts = re.split(r"/|\\", filename)[-1].split(".")
        if len(parts) > 1:
            ext = "." + parts.pop()
            filename = filename[: -len(ext)]
        else:
            ext = ""
        if filename == "":
            filename = "__"
        if len(ext) > 254:
            ext = ext[254:]
        maxl = 255 - len(ext)
        filename = filename[:maxl]
        filename = filename + ext
        # Re-check last character (if there was no extension)
        filename = filename.rstrip(". ")
        if len(filename) == 0:
            filename = "__"
    return filename



#!/usr/bin/env python3
from functools import partial
from pathlib import Path
from subprocess import Popen
from tkinter import filedialog
from typing import Any

from ttkbootstrap import Button, Entry, Frame, Label  # type: ignore

from sticker_convert.gui_components.gui_utils import GUIUtils
from sticker_convert.gui_components.windows.base_window import BaseWindow
from sticker_convert.utils.auth.get_signal_auth import GetSignalAuth


class SignalGetAuthWindow(BaseWindow):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.title("Get Signal uuid and password")

        self.cb_msg_block_signal = partial(self.gui.cb_msg_block, parent=self)
        self.cb_ask_str_signal = partial(self.gui.cb_ask_str, parent=self)

        self.frame_info = Frame(self.scrollable_frame)
        self.frame_btns = Frame(self.scrollable_frame)
        self.frame_config = Frame(self.scrollable_frame)

        self.frame_info.grid(column=0, row=0, sticky="news", padx=3, pady=3)
        self.frame_btns.grid(column=0, row=1, sticky="news", padx=3, pady=3)
        self.frame_config.grid(column=0, row=2, sticky="news", padx=3, pady=3)

        # Info frame
        self.explanation_lbl = Label(
            self.frame_info,
            text="Please install Signal Desktop and login first.",
            justify="left",
            anchor="w",
        )

        self.explanation_lbl.grid(column=0, row=0, sticky="w", padx=3, pady=3)

        # Start button frame
        self.launch_btn = Button(
            self.frame_btns, text="Launch Signal Desktop", command=self.cb_launch_signal
        )

        self.login_btn = Button(
            self.frame_btns, text="Get uuid and password", command=self.cb_login
        )

        self.launch_btn.pack()
        self.login_btn.pack()

        # Config frame
        self.setdir_lbl = Label(
            self.frame_config,
            text=self.gui.help["cred"]["signal_data_dir"],
            justify="left",
            anchor="w",
        )

        self.setdir_entry = Entry(
            self.frame_config,
            textvariable=self.gui.signal_data_dir_var,
            width=32,
        )
        self.setdir_btn = Button(
            self.frame_config,
            text="Choose",
            command=self.cb_setdir,
            width=8,
            bootstyle="secondary",  # type: ignore
        )

        self.setdir_lbl.grid(column=0, row=0, columnspan=2, sticky="w", padx=3, pady=3)
        self.setdir_entry.grid(column=0, row=1, sticky="w", padx=3, pady=3)
        self.setdir_btn.grid(column=1, row=1, sticky="e", padx=3, pady=3)

        GUIUtils.finalize_window(self)

    def cb_login(self) -> None:
        m = GetSignalAuth()

        signal_bin_path = None
        signal_user_data_dir = None
        if self.gui.signal_data_dir_var.get():
            signal_bin_path = "(User specified)"
            signal_user_data_dir = self.gui.signal_data_dir_var.get()

        uuid, password, msg = m.get_cred(signal_bin_path, signal_user_data_dir)

        if uuid and password:
            if not self.gui.creds.get("signal"):
                self.gui.creds["signal"] = {}
            self.gui.creds["signal"]["uuid"] = uuid
            self.gui.creds["signal"]["password"] = password
            self.gui.signal_uuid_var.set(uuid)
            self.gui.signal_password_var.set(password)

            self.gui.save_creds()
            self.gui.highlight_fields()

        self.cb_msg_block_signal(msg)

    def cb_launch_signal(self) -> None:
        m = GetSignalAuth()
        signal_bin_path, signal_user_data_dir = m.get_signal_desktop()

        if self.gui.signal_data_dir_var.get():
            signal_user_data_dir = self.gui.signal_data_dir_var.get()

        if signal_bin_path:
            Popen(
                [
                    signal_bin_path,
                    "--no-sandbox",
                    f"--user-data-dir={signal_user_data_dir}",
                ]
            )
        else:
            self.cb_msg_block_signal("Error: Signal Desktop not installed.")

    def cb_setdir(self) -> None:
        orig_input_dir = self.gui.signal_data_dir_var.get()
        if not Path(orig_input_dir).is_dir():
            orig_input_dir = ""
        input_dir = filedialog.askdirectory(initialdir=orig_input_dir)
        if input_dir:
            self.gui.signal_data_dir_var.set(input_dir)



import os
from pathlib import Path
from typing import List

import pytest
from _pytest._py.path import LocalPath

from tests.common import KAKAO_TOKEN, LINE_COOKIES, PYTHON_EXE, SRC_DIR, TELEGRAM_TOKEN, run_cmd

TEST_DOWNLOAD = os.environ.get("TEST_DOWNLOAD")


def _run_sticker_convert(
    tmp_path: LocalPath,
    source: str,
    url: str,
    expected_file_count: int,
    expected_file_formats: List[str],
    with_title: bool,
    with_author: bool,
    with_emoji: bool,
):
    input_dir = Path(tmp_path) / "input"
    output_dir = Path(tmp_path) / "output"

    run_cmd(
        cmd=[
            PYTHON_EXE,
            "sticker-convert.py",
            f"--download-{source}",
            url,
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "--no-confirm",
            "--author",
            "sticker-convert-test",
            "--title",
            "sticker-convert-test",
        ],
        cwd=SRC_DIR,
    )

    for i in range(expected_file_count):
        for fmt in expected_file_formats:
            fname = Path(str(i).zfill(3) + fmt)
            assert fname in input_dir.iterdir()

    if with_title:
        assert Path("title.txt") in input_dir.iterdir()
    if with_author:
        assert Path("author.txt") in input_dir.iterdir()
    if with_emoji:
        assert Path("emoji.txt") in input_dir.iterdir()


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_signal_static_png(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="signal",
        url="https://signal.art/addstickers/#pack_id=caf8b92fcadce4b7f33ad949f2d8754b&pack_key=decb0e6cdec7683ee2fe54a291aa1e50db19dbab4a063f0cb1610aeda146698c",
        expected_file_count=3,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_signal_static_webp(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="signal",
        url="https://signal.art/addstickers/#pack_id=8c865ab7218386ceddbb563681634e22&pack_key=0394dfca57e10a34ea70cad834e343d90831c9f69164b03c813575c34873ef8d",
        expected_file_count=3,
        expected_file_formats=[".webp"],
        with_title=True,
        with_author=True,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_signal_animated_apng(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="signal",
        url="https://signal.art/addstickers/#pack_id=842af30fbf4dc7a502dbe15385e6ceb6&pack_key=4a5d6a5de108dc4eb873d900bf31a70ac312046751475a6c42195dbf7b729d48",
        expected_file_count=3,
        expected_file_formats=[".apng"],
        with_title=True,
        with_author=True,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_telegram_static_webp(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="telegram",
        url="https://t.me/addstickers/sticker_convert_test_by_laggykillerstickerbot",
        expected_file_count=3,
        expected_file_formats=[".webp"],
        with_title=True,
        with_author=False,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
@pytest.mark.skipif(not TELEGRAM_TOKEN, reason="No credentials")
def test_download_telegram_animated_webm(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="telegram",
        url="https://t.me/addstickers/sticker_convert_test_animated_by_laggykillerstickerbot",
        expected_file_count=3,
        expected_file_formats=[".webp"],
        with_title=True,
        with_author=False,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
@pytest.mark.skipif(not TELEGRAM_TOKEN, reason="No credentials")
def test_download_telegram_animated_tgs(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="telegram",
        url="https://telegram.me/addstickers/ColoredCats",
        expected_file_count=30,
        expected_file_formats=[".tgs"],
        with_title=True,
        with_author=False,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
@pytest.mark.skipif(not TELEGRAM_TOKEN, reason="No credentials")
def test_download_telegram_emoji(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="telegram",
        url="https://t.me/addemoji/ragemojis",
        expected_file_count=39,
        expected_file_formats=[".webp"],
        with_title=True,
        with_author=False,
        with_emoji=True,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_below_775(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/1/en",
        expected_file_count=88,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_no_region_lock(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/26407/",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_with_region_lock(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/12320864/zh-Hant",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_officialaccount(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/officialaccount/event/sticker/27404/zh-Hant",
        expected_file_count=16,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_officialaccount_special(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/officialaccount/event/sticker/27239/ja",
        expected_file_count=16,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/8831/",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_sound(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/5440/ja",
        expected_file_count=24,
        expected_file_formats=[".png", ".m4a"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_popup(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/22229788/ja",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_popup_foreground(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/14011294/ja",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_name_text(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/13780/ja",
        expected_file_count=40,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_per_sticker_text_no_cookies(
    tmp_path: LocalPath,
) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/12188389/ja",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
@pytest.mark.skipif(not LINE_COOKIES, reason="No credentials")
def test_download_line_animated_apng_per_sticker_text_with_cookies(
    tmp_path: LocalPath,
) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/stickershop/product/12188389/ja",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_static_png_emoji(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/emojishop/product/5f290aa26f7dd32fa145906b/ja",
        expected_file_count=40,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_line_animated_apng_emoji(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="line",
        url="https://store.line.me/emojishop/product/6124aa4ae72c607c18108562/ja",
        expected_file_count=40,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_kakao_static_png(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="kakao",
        url="https://e.kakao.com/t/pretty-all-friends",
        expected_file_count=32,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_kakao_animated_gif_store_link_no_token(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="kakao",
        url="https://e.kakao.com/t/lovey-dovey-healing-bear",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
@pytest.mark.skipif(not KAKAO_TOKEN, reason="No credentials")
def test_download_kakao_animated_gif_store_link_with_token(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="kakao",
        url="https://e.kakao.com/t/lovey-dovey-healing-bear",
        expected_file_count=24,
        expected_file_formats=[".webp"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )


@pytest.mark.skipif(not TEST_DOWNLOAD, reason="TEST_DOWNLOAD not set")
def test_download_kakao_animated_gif_share_link(tmp_path: LocalPath) -> None:
    _run_sticker_convert(
        tmp_path=tmp_path,
        source="kakao",
        url="https://emoticon.kakao.com/items/lV6K2fWmU7CpXlHcP9-ysQJx9rg=?referer=share_link",
        expected_file_count=24,
        expected_file_formats=[".png"],
        with_title=True,
        with_author=True,
        with_emoji=False,
    )



#!/usr/bin/env python3
import string
from typing import Optional
from urllib.parse import urlparse


class UrlDetect:
    @staticmethod
    def detect(url: str) -> Optional[str]:
        domain = urlparse(url).netloc

        if domain == "signal.art":
            return "signal"

        if domain in ("telegram.me", "t.me"):
            return "telegram"

        if (
            domain in ("store.line.me", "line.me")
            or url.startswith("line://shop/detail/")
            or (len(url) == 24 and all(c in string.hexdigits for c in url))
        ):
            return "line"

        if domain in ("e.kakao.com", "emoticon.kakao.com") or url.startswith(
            "kakaotalk://store/emoticon/"
        ):
            return "kakao"

        return None



#!/usr/bin/env python3
from pathlib import Path
from tkinter import filedialog
from typing import TYPE_CHECKING, Any

from ttkbootstrap import Button, Entry, Label, LabelFrame, OptionMenu  # type: ignore

from sticker_convert.definitions import DEFAULT_DIR
from sticker_convert.gui_components.frames.right_clicker import RightClicker
from sticker_convert.utils.url_detect import UrlDetect

if TYPE_CHECKING:
    from sticker_convert.gui import GUI  # type: ignore


class InputFrame(LabelFrame):
    def __init__(self, gui: "GUI", *args: Any, **kwargs: Any) -> None:
        self.gui = gui
        super().__init__(*args, **kwargs)

        self.input_option_lbl = Label(
            self, text="Input source", width=15, justify="left", anchor="w"
        )
        input_full_names = [i["full_name"] for i in self.gui.input_presets.values()]
        default_input_full_name = self.gui.input_presets[self.gui.default_input_mode][
            "full_name"
        ]
        self.input_option_opt = OptionMenu(
            self,
            self.gui.input_option_display_var,
            default_input_full_name,
            *input_full_names,
            command=self.cb_input_option,
            bootstyle="secondary",  # type: ignore
        )
        self.input_option_opt.config(width=32)

        self.input_setdir_lbl = Label(
            self, text="Input directory", width=35, justify="left", anchor="w"
        )
        self.input_setdir_entry = Entry(
            self,
            textvariable=self.gui.input_setdir_var,
            width=60,
            validatecommand=self.gui.highlight_fields,
        )
        self.input_setdir_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.setdir_btn = Button(
            self,
            text="Choose directory",
            command=self.cb_set_indir,
            width=16,
            bootstyle="secondary",  # type: ignore
        )

        self.address_lbl = Label(
            self,
            text=self.gui.input_presets[self.gui.default_input_mode]["address_lbls"],
            width=18,
            justify="left",
            anchor="w",
        )
        self.address_entry = Entry(
            self,
            textvariable=self.gui.input_address_var,
            width=80,
            validate="focusout",
            validatecommand=self.cb_input_option,
        )
        self.address_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.address_tip = Label(
            self,
            text=self.gui.input_presets[self.gui.default_input_mode]["example"],
            justify="left",
            anchor="w",
        )

        self.input_option_lbl.grid(column=0, row=0, sticky="w", padx=3, pady=3)
        self.input_option_opt.grid(
            column=1, row=0, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.input_setdir_lbl.grid(
            column=0, row=1, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.input_setdir_entry.grid(column=1, row=1, sticky="w", padx=3, pady=3)
        self.setdir_btn.grid(column=2, row=1, sticky="e", padx=3, pady=3)
        self.address_lbl.grid(column=0, row=2, sticky="w", padx=3, pady=3)
        self.address_entry.grid(
            column=1, row=2, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.address_tip.grid(column=0, row=3, columnspan=3, sticky="w", padx=3, pady=3)

        preset = [
            k
            for k, v in self.gui.input_presets.items()
            if v["full_name"] == self.gui.input_option_display_var.get()
        ][0]
        if preset == "local":
            self.address_entry.config(state="disabled")
        else:
            self.address_entry.config(state="normal")

    def cb_set_indir(self, *_args: Any) -> None:
        orig_input_dir = self.gui.input_setdir_var.get()
        if not Path(orig_input_dir).is_dir():
            orig_input_dir = DEFAULT_DIR
        input_dir = filedialog.askdirectory(initialdir=orig_input_dir)
        if input_dir:
            self.gui.input_setdir_var.set(input_dir)

    def cb_input_option(self, *_: Any) -> bool:
        input_option_display = self.gui.get_input_display_name()

        if input_option_display == "auto":
            url = self.gui.input_address_var.get()
            download_option = UrlDetect.detect(url)

            if download_option is None:
                self.gui.input_option_true_var.set(
                    self.gui.input_presets["auto"]["full_name"]
                )
            else:
                self.gui.input_option_true_var.set(
                    self.gui.input_presets[download_option]["full_name"]
                )
        else:
            self.gui.input_option_true_var.set(self.gui.input_option_display_var.get())

        self.gui.highlight_fields()

        return True

    def set_states(self, state: str) -> None:
        self.input_option_opt.config(state=state)
        self.address_entry.config(state=state)
        self.input_setdir_entry.config(state=state)
        self.setdir_btn.config(state=state)



#!/usr/bin/env python3
import copy
import re
from pathlib import Path
from queue import Queue
from typing import Any, Dict, List, Optional, Union, cast

import anyio
from telegram import InputFile, InputSticker, Sticker
from telegram.error import TelegramError
from telegram.ext import AIORateLimiter, ApplicationBuilder

from sticker_convert.converter import StickerConvert
from sticker_convert.job_option import CompOption, CredOption, OutputOption
from sticker_convert.uploaders.upload_base import UploadBase
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.metadata_handler import MetadataHandler
from sticker_convert.utils.media.format_verify import FormatVerify


class UploadTelegram(UploadBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.base_spec.size_max_img = 512000
        self.base_spec.size_max_vid = 256000
        self.base_spec.square = True
        self.base_spec.duration_max = 3000
        self.base_spec.set_res(512)

        self.png_spec = copy.deepcopy(self.base_spec)
        self.png_spec.set_format((".png",))
        self.png_spec.animated = False

        self.tgs_spec = copy.deepcopy(self.base_spec)
        self.tgs_spec.set_format((".tgs",))
        self.tgs_spec.fps_min = 60
        self.tgs_spec.fps_max = 60
        self.tgs_spec.size_max_img = 64000
        self.tgs_spec.size_max_vid = 64000

        self.webm_spec = copy.deepcopy(self.base_spec)
        self.webm_spec.set_format((".webm",))
        self.webm_spec.fps_max = 30
        self.webm_spec.animated = None if self.opt_comp.fake_vid else True

        self.opt_comp_merged = copy.deepcopy(self.opt_comp)
        self.opt_comp_merged.merge(self.base_spec)

        base_cover_spec = CompOption(
            size_max_img=128000, size_max_vid=32000, square=True, duration_max=3000
        )
        base_cover_spec.set_res(100)

        self.png_cover_spec = copy.deepcopy(base_cover_spec)
        self.png_cover_spec.set_format((".png",))
        self.png_cover_spec.animated = False

        self.tgs_cover_spec = copy.deepcopy(base_cover_spec)
        self.tgs_cover_spec.set_format((".tgs",))
        self.tgs_cover_spec.fps_min = 60
        self.tgs_cover_spec.fps_max = 60

        self.webm_cover_spec = copy.deepcopy(base_cover_spec)
        self.webm_cover_spec.set_format((".webm",))
        self.webm_cover_spec.fps_max = 30
        self.webm_cover_spec.animated = True

        self.opt_comp_cover_merged = copy.deepcopy(self.opt_comp)
        self.opt_comp_cover_merged.merge(self.base_spec)

    async def upload_pack(
        self, pack_title: str, stickers: List[Path], emoji_dict: Dict[str, str]
    ) -> Optional[str]:
        token = self.opt_cred.telegram_token.strip()
        assert token
        timeout = 10 * len(stickers)

        application = (  # type: ignore
            ApplicationBuilder()
            .token(self.opt_cred.telegram_token.strip())
            .rate_limiter(AIORateLimiter())
            .connect_timeout(timeout)
            .pool_timeout(timeout)
            .read_timeout(timeout)
            .write_timeout(timeout)
            .connection_pool_size(len(stickers))
            .build()
        )

        async with application:
            bot = application.bot
            pack_short_name = (
                pack_title.replace(" ", "_") + "_by_" + bot.name.replace("@", "")
            )
            pack_short_name = re.sub(
                "[^0-9a-zA-Z]+", "_", pack_short_name
            )  # name used in url, only alphanum and underscore only

            sticker_set = None
            try:
                sticker_set = await bot.get_sticker_set(
                    pack_short_name,
                    read_timeout=30,
                    write_timeout=30,
                    connect_timeout=30,
                    pool_timeout=30,
                )
            except TelegramError:
                pass

            if sticker_set is not None:
                question = f"Warning: Pack {pack_short_name} already exists.\n"
                question += "Delete all stickers in pack?\n"
                question += "Note: After recreating set, please wait for about 3 minutes for the set to reappear."

                self.cb.put(
                    (
                        "ask_bool",
                        (question,),
                        None,
                    )
                )
                if self.cb_return:
                    response = self.cb_return.get_response()
                else:
                    response = False

                if response is True:
                    self.cb.put(f"Deleting all stickers from pack {pack_short_name}")
                    await bot.delete_sticker_set(pack_short_name)
                    sticker_set = None
                else:
                    self.cb.put(f"Not deleting existing pack {pack_short_name}")

            if self.opt_output.option == "telegram_emoji":
                sticker_type = Sticker.CUSTOM_EMOJI
            else:
                sticker_type = Sticker.REGULAR

            input_stickers: List[InputSticker] = []
            sticker_format = None
            cover_spec_choice = None
            ext = None
            for src in stickers:
                self.cb.put(f"Verifying {src} for uploading to telegram")

                emoji = emoji_dict.get(Path(src).stem, None)
                if emoji:
                    if len(emoji) > 20:
                        self.cb.put(
                            f"Warning: {len(emoji)} emoji for file {Path(src).name}, exceeding limit of 20, keep first 20 only..."
                        )
                    emoji_list = [*emoji][:20]
                else:
                    self.cb.put(
                        f"Warning: Cannot find emoji for file {Path(src).name}, skip uploading this file..."
                    )
                    continue

                ext = Path(src).suffix
                if ext == ".tgs":
                    spec_choice = self.tgs_spec
                    cover_spec_choice = self.tgs_cover_spec
                    sticker_format = "animated"
                elif ext == ".webm":
                    spec_choice = self.webm_spec
                    cover_spec_choice = self.webm_cover_spec
                    sticker_format = "video"
                else:
                    ext = ".png"
                    spec_choice = self.png_spec
                    cover_spec_choice = self.png_cover_spec
                    sticker_format = "static"

                if self.opt_output.option == "telegram_emoji":
                    spec_choice.set_res(100)

                if FormatVerify.check_file(src, spec=spec_choice):
                    with open(src, "rb") as f:
                        sticker_bytes = f.read()
                else:
                    _, _, convert_result, _ = StickerConvert.convert(
                        Path(src),
                        Path(f"bytes{ext}"),
                        self.opt_comp_merged,
                        self.cb,
                        self.cb_return,
                    )
                    sticker_bytes = cast(bytes, convert_result)

                input_stickers.append(
                    InputSticker(sticker=sticker_bytes, emoji_list=emoji_list)
                )

            cover_path = MetadataHandler.get_cover(self.opt_output.dir)
            thumbnail_bytes: Union[None, bytes, Path] = None
            if cover_path and cover_spec_choice and ext:
                if FormatVerify.check_file(cover_path, spec=cover_spec_choice):
                    with open(cover_path, "rb") as f:
                        thumbnail_bytes = f.read()
                else:
                    _, _, thumbnail_bytes, _ = StickerConvert.convert(
                        cover_path,
                        Path(f"bytes{ext}"),
                        self.opt_comp_cover_merged,
                        self.cb,
                        self.cb_return,
                    )

            base_num = 1
            if sticker_set is None and sticker_format is not None:
                init_stickers = input_stickers[:50]
                if len(input_stickers) > 50:
                    amount_str = "first 50"
                else:
                    amount_str = "all"
                start_msg = f"Creating pack and bulk uploading {amount_str} stickers of {pack_short_name}"
                finish_msg = f"Created pack and bulk uploaded {amount_str} stickers of {pack_short_name}"
                error_msg = f"Cannot create pack and bulk upload {amount_str} stickers of {pack_short_name} due to"
                self.cb.put(start_msg)
                try:
                    await bot.create_new_sticker_set(
                        user_id=self.opt_cred.telegram_userid,
                        name=pack_short_name,
                        title=pack_title,
                        stickers=init_stickers,
                        sticker_format=sticker_format,
                        sticker_type=sticker_type,
                    )
                    self.cb.put(finish_msg)
                    input_stickers = input_stickers[50:]
                    base_num = 51
                except TelegramError as e:
                    self.cb.put(f"{error_msg} {e}")
                    return None

            if len(input_stickers) > 0:
                self.cb.put(f"Uploading stickers of {pack_short_name} one-by-one")
                for i, sticker in enumerate(input_stickers):
                    assert isinstance(sticker.sticker, InputFile)
                    num = base_num + i
                    try:
                        # We could use tg.start_soon() here
                        # But this would disrupt the order of stickers
                        await bot.add_sticker_to_set(
                            user_id=self.opt_cred.telegram_userid,
                            name=pack_short_name,
                            sticker=sticker,
                        )
                        self.cb.put(f"Uploaded sticker #{num} of {pack_short_name}")
                    except TelegramError as e:
                        self.cb.put(
                            f"Cannot upload sticker #{num} of {pack_short_name} due to {e}"
                        )

            if thumbnail_bytes is not None:
                try:
                    self.cb.put(
                        f"Uploading cover (thumbnail) of pack {pack_short_name}"
                    )
                    await bot.set_sticker_set_thumbnail(
                        name=pack_short_name,
                        user_id=self.opt_cred.telegram_userid,
                        thumbnail=thumbnail_bytes,
                    )
                    self.cb.put(f"Uploaded cover (thumbnail) of pack {pack_short_name}")
                except TelegramError as e:
                    self.cb.put(
                        f"Cannot upload cover (thumbnail) of pack {pack_short_name} due to {e}"
                    )

            self.cb.put(f"Finish uploading {pack_short_name}")

        if self.opt_output.option == "telegram_emoji":
            result = f"https://t.me/addemoji/{pack_short_name}"
        else:
            result = f"https://t.me/addstickers/{pack_short_name}"
        return result

    def upload_stickers_telegram(self) -> List[str]:
        urls: List[str] = []

        if not (self.opt_cred.telegram_token and self.opt_cred.telegram_userid):
            self.cb.put("Token and userid required for uploading to telegram")
            return urls

        title, _, emoji_dict = MetadataHandler.get_metadata(
            self.opt_output.dir,
            title=self.opt_output.title,
            author=self.opt_output.author,
        )
        if title is None:
            raise TypeError("title cannot be", title)
        if emoji_dict is None:
            msg_block = "emoji.txt is required for uploading signal stickers\n"
            msg_block += f"emoji.txt generated for you in {self.opt_output.dir}\n"
            msg_block += f"Default emoji is set to {self.opt_comp.default_emoji}.\n"
            msg_block += "Please edit emoji.txt now, then continue"
            MetadataHandler.generate_emoji_file(
                directory=self.opt_output.dir, default_emoji=self.opt_comp.default_emoji
            )

            self.cb.put(("msg_block", (msg_block,), None))
            if self.cb_return:
                self.cb_return.get_response()

            title, _, emoji_dict = MetadataHandler.get_metadata(
                self.opt_output.dir,
                title=self.opt_output.title,
                author=self.opt_output.author,
            )

        assert title is not None
        packs = MetadataHandler.split_sticker_packs(
            self.opt_output.dir,
            title=title,
            file_per_anim_pack=50,
            file_per_image_pack=120,
            separate_image_anim=not self.opt_comp.fake_vid,
        )

        for pack_title, stickers in packs.items():
            self.cb.put(f"Uploading pack {pack_title}")
            result = anyio.run(self.upload_pack, pack_title, stickers, emoji_dict)
            if result:
                self.cb.put((result))
                urls.append(result)

        return urls

    @staticmethod
    def start(
        opt_output: OutputOption,
        opt_comp: CompOption,
        opt_cred: CredOption,
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> List[str]:
        exporter = UploadTelegram(
            opt_output,
            opt_comp,
            opt_cred,
            cb,
            cb_return,
        )
        return exporter.upload_stickers_telegram()
